#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import unittest

import torch

try:
    from fast_transformers.clustering.hamming import cluster_cuda
except ImportError:
    pass


class TestClusterGPU(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not torch.cuda.is_available():
            raise unittest.SkipTest("No CUDA capable device detected")

    def test_long_clusters(self):
        for bits in range(1, 63):
            hashes = torch.cat([
                torch.zeros(50).long(),
                torch.ones(50).long() * (2**bits - 1)
            ]).view(1, 1, 100)[:,:,torch.randperm(100)].cuda()
            lengths = torch.full((1,), 100, dtype=torch.int32).cuda()
            centroids = torch.empty(1, 1, 2, dtype=torch.int64).cuda()
            distances = torch.empty(1, 1, 100, dtype=torch.int32).cuda()
            bitcounts = torch.empty(1, 1, 2, bits, dtype=torch.int32).cuda()
            clusters = torch.empty(1, 1, 100, dtype=torch.int32).cuda()
            counts = torch.empty(1, 1, 2, dtype=torch.int32).cuda()

            cluster_cuda.cluster(
                hashes,
                lengths,
                centroids,
                distances,
                bitcounts,
                clusters,
                counts,
                10,
                bits
            )
            self.assertEqual(
                tuple(sorted(centroids.cpu().numpy().ravel().tolist())),
                (0, 2**bits - 1)
            )
            self.assertTrue(torch.all(counts==50))

    def test_two_clusters(self):
        hashes = torch.cat([
            torch.zeros(50).long(),
            torch.full((50,), 255, dtype=torch.int64)
        ]).view(1, 1, 100)[:, :, torch.randperm(100)].cuda()
        lengths = torch.full((1,), 100, dtype=torch.int32).cuda()
        centroids = torch.empty(1, 1, 2, dtype=torch.int64).cuda()
        distances = torch.empty(1, 1, 100, dtype=torch.int32).cuda()
        bitcounts = torch.empty(1, 1, 2, 8, dtype=torch.int32).cuda()
        clusters = torch.empty(1, 1, 100, dtype=torch.int32).cuda()
        counts = torch.empty(1, 1, 2, dtype=torch.int32).cuda()

        cluster_cuda.cluster(
            hashes,
            lengths,
            centroids,
            distances,
            bitcounts,
            clusters,
            counts,
            10,
            8
        )
        self.assertEqual(
            tuple(sorted(centroids.cpu().numpy().ravel().tolist())),
            (0, 255)
        )
        self.assertTrue(torch.all(counts==50))

    def test_power_of_2_clusters(self):
        hashes = torch.cat([
            torch.full((10,), 1<<i, dtype=torch.int64)
            for i in range(8)
        ]).view(1, 1, 80)[:, :, torch.randperm(80)].cuda()
        lengths = torch.full((1,), 80, dtype=torch.int32).cuda()
        centroids = torch.empty(1, 1, 8, dtype=torch.int64).cuda()
        distances = torch.empty(1, 1, 80, dtype=torch.int32).cuda()
        bitcounts = torch.empty(1, 1, 8, 8, dtype=torch.int32).cuda()
        clusters = torch.empty(1, 1, 80, dtype=torch.int32).cuda()
        counts = torch.empty(1, 1, 8, dtype=torch.int32).cuda()

        cluster_cuda.cluster(
            hashes,
            lengths,
            centroids,
            distances,
            bitcounts,
            clusters,
            counts,
            2000,
            8
        )
        self.assertEqual(
            tuple(sorted(centroids.cpu().numpy().ravel().tolist())),
            (1, 2, 4, 8, 16, 32, 64, 128)
        )
        self.assertTrue(torch.all(counts==10))

    def test_many_sequences(self):
        hashes = torch.cat([
            torch.zeros(50).long(),
            torch.full((50,), 255, dtype=torch.int64)
        ]).view(1, 1, 100)[:, :, torch.randperm(100)].repeat(5, 3, 1).cuda()
        lengths = torch.full((5,), 100, dtype=torch.int32).cuda()
        centroids = torch.empty(5, 3, 2, dtype=torch.int64).cuda()
        distances = torch.empty(5, 3, 100, dtype=torch.int32).cuda()
        bitcounts = torch.empty(5, 3, 2, 8, dtype=torch.int32).cuda()
        clusters = torch.empty(5, 3, 100, dtype=torch.int32).cuda()
        counts = torch.empty(5, 3, 2, dtype=torch.int32).cuda()

        cluster_cuda.cluster(
            hashes,
            lengths,
            centroids,
            distances,
            bitcounts,
            clusters,
            counts,
            10,
            8
        )
        self.assertTrue(torch.all(centroids.min(-1)[0] == 0))
        self.assertTrue(torch.all(centroids.max(-1)[0] == 255))
        self.assertTrue(torch.all(counts==50))


if __name__ == "__main__":
    unittest.main()




#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import os
import time
import unittest

import torch
from torch.nn.init import normal_

from fast_transformers.aggregate import aggregate, broadcast
from fast_transformers.hashing import compute_hashes
from fast_transformers.clustering.hamming import cluster
from fast_transformers.sparse_product import sparse_dot_product
from fast_transformers.sparse_product import clustered_sparse_dot_product

def cluster_queries(Q, query_lengths, C, I, B):
    N, H, L, E = Q.shape
    planes = Q.new_empty((B, E+1))
    normal_(planes)
    planes[:, -1] = 0
    hashes = compute_hashes(Q.view(N*H*L, E), planes).view(N, H, L)
    # Cluster the hashes and return the cluster index per query
    groups, counts =  cluster(
        hashes,
        query_lengths,
        clusters=C,
        iterations=I,
        bits=B
    )

    return groups, counts


class TestSparseProductCUDA(unittest.TestCase):
    @property
    def device(self):
        return "cuda"

    @classmethod
    def setUpClass(cls):
        if not torch.cuda.is_available():
            raise unittest.SkipTest("No CUDA capable device detected")

    def test_simple_product(self):
        N = 2
        H = 2
        L = 1000
        E = 32
        S = 1000
        k = 32
        C = 50
        I = 5
        B = 16

        Q = torch.randn(N, H, L, E).to(self.device)
        K = torch.randn(N, H, S, E).to(self.device)
        lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
        groups, counts = cluster_queries(Q, lengths, C, I, B)
        Q_grouped = aggregate(Q, groups, 1/counts.float())
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
        _, topk = torch.topk(QK, k, dim=-1)
        topk = topk.contiguous()
        products = clustered_sparse_dot_product(Q, K, topk, groups, counts, lengths)
        topk_broadcast = broadcast(
            topk.float(),
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, k), device=Q.device)
        )

        all_products = torch.einsum("nhle,nhse->nhls", Q, K)
        products_2 = all_products[
            torch.arange(N).view(N, 1, 1, 1),
            torch.arange(H).view(1, H, 1, 1),
            torch.arange(L).view(1, 1, L, 1),
            topk_broadcast.long()
        ]

        self.assertLess(
            torch.max(torch.abs(
                products_2 - products
                )
            ),
            1e-4
        )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_small_benchmark(self):
        N = 12
        H = 8
        L = 1000
        E = 32
        S = 1000
        k = 32
        C = 100
        I = 10
        B = 32

        Q = torch.randn(N, H, L, E).to(self.device)
        K = torch.randn(N, H, S, E).to(self.device)
        lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
        groups, counts = cluster_queries(Q, lengths, C, I, B)
        Q_grouped = aggregate(Q, groups, 1/counts.float())
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
        _, topk = torch.topk(QK, k, dim=-1)
        topk = topk.contiguous()
        products = torch.zeros((N, H, L, k), dtype=torch.float32).to(self.device)
        products = clustered_sparse_dot_product(Q, K, topk, groups, counts, lengths)
        for i in range(1000):
            products = clustered_sparse_dot_product(
                Q,
                K,
                topk,
                groups,
                counts,
                lengths
            )

        torch.cuda.synchronize()
        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        products = clustered_sparse_dot_product(
            Q,
            K,
            topk,
            groups,
            counts,
            lengths
        )
        e.record()
        torch.cuda.synchronize()
        t_sc = s.elapsed_time(e)
   
        topk_broadcast = broadcast(
            topk.float(),
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, k), device=Q.device)
        )

        for i in range(1000):
            products = sparse_dot_product(
                Q,
                K,
                topk_broadcast.long()
            )
        torch.cuda.synchronize()
        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        products_s = sparse_dot_product(
            Q,
            K,
            topk_broadcast.long(),
        )
        e.record()
        torch.cuda.synchronize()
        t_s = s.elapsed_time(e)

        for i in range(1000):
            torch.einsum("nhle,nhse->nhls", Q, K)
        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        torch.einsum("nhle,nhse->nhls", Q, K)
        e.record()
        torch.cuda.synchronize()
        t_f = s.elapsed_time(e)
        print("Sparse_Clustered: {}, Sparse: {}, Full: {}".format(t_sc, t_s, t_f))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#


class Registry(object):
    """Hold the available attention implementations and their required
    parameters."""
    def __init__(self):
        self._classes = {}
        self._class_params = {}
        self._parameters = {}

    def register(self, key, class_object, parameter_tuples):
        # register the class if the key is new
        if key in self._classes:
            raise ValueError("{} is already registered".format(key))
        self._classes[key] = class_object

        # register the parameters
        for parameter, spec in parameter_tuples:
            if (
                parameter in self._parameters and
                self._parameters[parameter] != spec
            ):
                raise ValueError(("{} is already registered with "
                                  "spec {!r} instead of {!r}").format(
                                  parameter,
                                  self._parameters[parameter],
                                  spec
                                ))
            self._parameters[parameter] = spec

        # note which parameters are needed by this class
        self._class_params[key] = [p for p, s in parameter_tuples]

    def __contains__(self, key):
        return key in self._classes

    def __getitem__(self, key):
        return self._classes[key], self._class_params[key]

    @property
    def keys(self):
        return list(self._classes.keys())

    def contains_parameter(self, key):
        return key in self._parameters

    def validate_parameter(self, key, value):
        return self._parameters[key].get(value)


AttentionRegistry = Registry()
RecurrentAttentionRegistry = Registry()
RecurrentCrossAttentionRegistry = Registry()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""This module implements builders that simplify building complex transformer
architectures with different attention mechanisms.

The main idea is to facilitate the construction of various attention layers and
transformer encoder layers and simplify their assembly into one transformer
module. It also allows for flexibility in the scripts as many builder
parameters can correspond 1-1 with command line arguments.

Example usage:

    builder = TransformerEncoderBuilder()
    builder.n_layers = 12
    builder.n_heads = 8
    builder.feed_forward_dimensions = 1024
    builder.query_dimensions = 64
    builder.value_dimensions = 64
    builder.dropout = 0.1
    builder.attention_dropout = 0.1
    builder.attention_type = "linear"
    transformer = builder.get()
"""

__all__ = [
    "AttentionBuilder",
    "RecurrentAttentionBuilder",
    "RecurrentCrossAttentionBuilder"
]

# Import the attention implementations so that they register themselves with
# the builder. Attention implementations external to the library should be
# imported before using the builders.
#
# TODO: Should this behaviour change? Namely, should all attention
#       implementations be imported in order to be useable? This also allows
#       using the library even partially built, for instance.
from ..attention import \
    FullAttention, \
    LinearAttention, CausalLinearAttention, \
    ClusteredAttention, ImprovedClusteredAttention, \
    ReformerAttention, \
    ExactTopKAttention, ImprovedClusteredCausalAttention, \
    ConditionalFullAttention
del FullAttention, \
    LinearAttention, CausalLinearAttention, \
    ClusteredAttention, ImprovedClusteredAttention, \
    ReformerAttention, \
    ExactTopKAttention, ImprovedClusteredCausalAttention, \
    ConditionalFullAttention


from .attention_builders import \
    AttentionBuilder, \
    RecurrentAttentionBuilder, \
    RecurrentCrossAttentionBuilder

from .transformer_builders import \
    TransformerEncoderBuilder, \
    RecurrentEncoderBuilder, \
    TransformerDecoderBuilder, \
    RecurrentDecoderBuilder



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import torch

from .hash_cpu import compute_hashes as compute_hashes_cpu
try:
    from .hash_cuda import compute_hashes as compute_hashes_cuda
except ImportError:
    pass


def compute_hashes(X, A, H=None):
    device = X.device
    if H is None:
        H = torch.zeros(len(X), dtype=torch.int64, device=device)
    else:
        H.zero_()
    if A.shape[1] != X.shape[1] + 1:
        raise ValueError("The hash requires a bias")

    if device.type == "cpu":
        compute_hashes_cpu(X, A, H)
    else:
        compute_hashes_cuda(X, A, H)

    return H



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import unittest

import torch

from fast_transformers.recurrent.attention import RecurrentAttentionLayer


class TestRecurrentAttentionLayer(unittest.TestCase):
    def _assert_sizes_attention(self, qshape, kshape, vshape):
        def inner(q, k, v, m):
            self.assertEqual(q.shape, qshape)
            self.assertEqual(k.shape, kshape)
            self.assertEqual(v.shape, vshape)
            N, H, E = q.shape
            _, _, D = v.shape
            return v.new_zeros((N, H, D)), m
        return inner

    def test_forward(self):
        att = RecurrentAttentionLayer(
            self._assert_sizes_attention(
                (10, 4, 25),
                (10, 4, 25),
                (10, 4, 25)
            ),
            100,
            4
        )
        v, m = att(
            torch.rand(10, 100),
            torch.rand(10, 100),
            torch.rand(10, 100),
            "test memory"
        )
        self.assertEqual(v.shape, (10, 100))
        self.assertEqual(m, "test memory")

        att = RecurrentAttentionLayer(
            self._assert_sizes_attention(
                (10, 4, 32),
                (10, 4, 32),
                (10, 4, 64)
            ),
            100,
            4,
            d_keys=32,
            d_values=64
        )
        v, m = att(
            torch.rand(10, 100),
            torch.rand(10, 100),
            torch.rand(10, 100),
            "test memory"
        )
        self.assertEqual(v.shape, (10, 100))
        self.assertEqual(m, "test memory")


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""The base attention layer performs all the query key value projections and
output projections leaving the implementation of the attention to the inner
attention module.

The transformer layers, however, are agnostic of the attention implementation
and any layer that implements the same interface can substitute for the
attention layer.
"""

from torch.nn import Linear, Module


class AttentionLayer(Module):
    """Implement the attention layer. Namely project the inputs to multi-head
    queries, keys and values, call the attention implementation and then
    reproject the output.

    It can be thought of as a decorator (see decorator design patter) of an
    attention layer.

    Arguments
    ---------
        attention: Specific inner attention implementation that just computes a
                   weighted average of values given a similarity of queries and
                   keys.
        d_model: The input feature dimensionality
        n_heads: The number of heads for the multi head attention
        d_keys: The dimensionality of the keys/queries
                (default: d_model/n_heads)
        d_values: The dimensionality of the values (default: d_model/n_heads)
    """
    def __init__(self, attention, d_model, n_heads, d_keys=None,
                 d_values=None):
        super(AttentionLayer, self).__init__()

        # Fill d_keys and d_values
        d_keys = d_keys or (d_model//n_heads)
        d_values = d_values or (d_model//n_heads)

        self.inner_attention = attention
        self.query_projection = Linear(d_model, d_keys * n_heads)
        self.key_projection = Linear(d_model, d_keys * n_heads)
        self.value_projection = Linear(d_model, d_values * n_heads)
        self.out_projection = Linear(d_values * n_heads, d_model)
        self.n_heads = n_heads

    def forward(self, queries, keys, values, attn_mask, query_lengths,
                key_lengths):
        """Apply attention to the passed in queries/keys/values after
        projecting them to multiple heads.

        In the argument description we make use of the following sizes

            - N: the batch size
            - L: The maximum length of the queries
            - S: The maximum length of the keys (the actual length per sequence
              is given by the length mask)
            - D: The input feature dimensionality passed in the constructor as
              'd_model'

        Arguments
        ---------
            queries: (N, L, D) The tensor containing the queries
            keys: (N, S, D) The tensor containing the keys
            values: (N, S, D) The tensor containing the values
            attn_mask: An implementation of BaseMask that encodes where each
                       query can attend to
            query_lengths: An implementation of  BaseMask that encodes how
                           many queries each sequence in the batch consists of
            key_lengths: An implementation of BaseMask that encodes how
                         many queries each sequence in the batch consists of

        Returns
        -------
            The new value for each query as a tensor of shape (N, L, D).
        """
        # Project the queries/keys/values
        queries = self.query_projection(queries)
        keys = self.key_projection(keys)
        values = self.value_projection(values)

        # Reshape them into many heads and compute the attention
        N, L, D = queries.shape
        _, S, _ = keys.shape
        H = self.n_heads
        new_values = self.inner_attention(
            queries.view(N, L, H, -1),
            keys.view(N, S, H, -1),
            values.view(N, S, H, -1),
            attn_mask,
            query_lengths,
            key_lengths
        ).view(N, L, -1)

        # Project the output and return
        return self.out_projection(new_values)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""Implement unmasked linear attention."""

import torch
from torch.nn import Module

from ..attention_registry import AttentionRegistry, Optional, Callable


def elu_feature_map(x):
    return torch.nn.functional.elu(x) + 1


class LinearAttention(Module):
    """Implement unmasked attention using dot product of feature maps in
    O(N D^2) complexity.

    Given the queries, keys and values as Q, K, V instead of computing

        V' = softmax(Q.mm(K.t()), dim=-1).mm(V),

    we make use of a feature map function Φ(.) and perform the following
    computation

        V' = normalize(Φ(Q).mm(Φ(K).t())).mm(V).

    The above can be computed in O(N D^2) complexity where D is the
    dimensionality of Q, K and V and N is the sequence length. Depending on the
    feature map, however, the complexity of the attention might be limited.

    Arguments
    ---------
        feature_map: callable, a callable that applies the feature map to the
                     last dimension of a tensor (default: elu(x)+1)
        eps: float, a small number to ensure the numerical stability of the
             denominator (default: 1e-6)
    """
    def __init__(self, feature_map=None, eps=1e-6):
        super(LinearAttention, self).__init__()
        self.feature_map = feature_map or elu_feature_map
        self.eps = eps

    def forward(self, queries, keys, values, attn_mask, query_lengths,
                key_lengths):
        # Apply the feature map to the queries and keys
        Q = self.feature_map(queries)
        K = self.feature_map(keys)

        # Apply the key padding mask and make sure that the attn_mask is
        # all_ones
        if not attn_mask.all_ones:
            raise RuntimeError(("LinearAttention does not support arbitrary "
                                "attention masks"))
        K = K * key_lengths.float_matrix[:, :, None, None]

        # Compute the KV matrix, namely the dot product of keys and values so
        # that we never explicitly compute the attention matrix and thus
        # decrease the complexity
        KV = torch.einsum("nshd,nshm->nhmd", K, values)

        # Compute the normalizer
        Z = 1/(torch.einsum("nlhd,nhd->nlh", Q, K.sum(dim=1))+self.eps)

        # Finally compute and return the new values
        V = torch.einsum("nlhd,nhmd,nlh->nlhm", Q, KV, Z)

        return V.contiguous()


# Register the attention implementation so that it becomes available in our
# builders
AttentionRegistry.register(
    "linear", LinearAttention,
    [("feature_map", Optional(Callable))]
)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import unittest
import os
import time

import torch

from fast_transformers.clustering.hamming import cluster_cpu
from fast_transformers.hashing import compute_hashes
from fast_transformers.clustering.hamming import cluster


def generate_hash(n_points, d, b):
    x = torch.rand(n_points, d)
    a = torch.randn(b, d+1)
    a[:,d] = 0
    h = torch.zeros(n_points, dtype=torch.int64)
    compute_hashes(x, a, h)
    return h


class TestClusterCPU(unittest.TestCase):
    def test_long_clusters(self):
        for bits in range(1, 63):
            hashes = torch.cat([
                torch.zeros(50).long(),
                torch.ones(50).long() * (2**bits - 1)
            ]).view(1, 1, 100)[:,:,torch.randperm(100)]
            lengths = torch.full((1,), 100, dtype=torch.int32)
            centroids = torch.empty(1, 1, 2, dtype=torch.int64)
            clusters = torch.empty(1, 1, 100, dtype=torch.int32)
            counts = torch.empty(1, 1, 2, dtype=torch.int32)

            cluster_cpu(
                hashes,
                lengths,
                centroids,
                clusters,
                counts,
                10,
                bits
            )
            self.assertEqual(
                tuple(sorted(centroids.numpy().ravel().tolist())),
                (0, 2**bits - 1)
            )
            self.assertTrue(torch.all(counts==50))

    def test_two_clusters(self):
        hashes = torch.cat([
            torch.zeros(50).long(),
            torch.full((50,), 255, dtype=torch.int64)
        ]).view(1, 1, 100)[:,:,torch.randperm(100)]
        lengths = torch.full((1,), 100, dtype=torch.int32)
        centroids = torch.empty(1, 1, 2, dtype=torch.int64)
        clusters = torch.empty(1, 1, 100, dtype=torch.int32)
        counts = torch.empty(1, 1, 2, dtype=torch.int32)

        cluster_cpu(
            hashes,
            lengths,
            centroids,
            clusters,
            counts,
            10,
            8
        )
        self.assertEqual(
            tuple(sorted(centroids.numpy().ravel().tolist())),
            (0, 255)
        )
        self.assertTrue(torch.all(counts==50))

    def test_power_of_2_clusters(self):
        hashes = torch.cat([
            torch.full((10,), 1<<i, dtype=torch.int64)
            for i in range(8)
        ]).view(1, 1, 80)[:,:,torch.randperm(80)]
        lengths = torch.full((1,), 80, dtype=torch.int32)
        centroids = torch.empty(1, 1, 8, dtype=torch.int64)
        clusters = torch.empty(1, 1, 80, dtype=torch.int32)
        counts = torch.empty(1, 1, 8, dtype=torch.int32)

        cluster_cpu(
            hashes,
            lengths,
            centroids,
            clusters,
            counts,
            2000,
            8
        )
        self.assertEqual(
            tuple(sorted(centroids.numpy().ravel().tolist())),
            (1, 2, 4, 8, 16, 32, 64, 128)
        )
        self.assertTrue(torch.all(counts==10))

    def test_many_sequences(self):
        hashes = torch.cat([
            torch.zeros(50).long(),
            torch.full((50,), 255, dtype=torch.int64)
        ]).view(1, 1, 100)[:,:,torch.randperm(100)].repeat(5, 3, 1)
        lengths = torch.full((5,), 100, dtype=torch.int32)
        centroids = torch.empty(5, 3, 2, dtype=torch.int64)
        clusters = torch.empty(5, 3, 100, dtype=torch.int32)
        counts = torch.empty(5, 3, 2, dtype=torch.int32)

        cluster_cpu(
            hashes,
            lengths,
            centroids,
            clusters,
            counts,
            10,
            8
        )
        self.assertTrue(torch.all(centroids.min(2)[0] == 0))
        self.assertTrue(torch.all(centroids.max(2)[0] == 255))
        self.assertTrue(torch.all(counts==50))

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_benchmark_clustering(self):
        N=12
        H=4 
        L=1000
        E=32 
        
        k=100
        n_buckets=63
        n_iterations=10
        
        n_points = L * N * H
        for n_buckets in range(10, 64):        
            hashes = generate_hash(n_points, E, n_buckets).view(N, H, L)
            groups = torch.zeros((N, H, L), dtype=torch.int32)
            counts = torch.zeros((N, H, k), dtype=torch.int32)
            centroids = torch.zeros((N, H, k), dtype=torch.int64)
            distances = torch.zeros((N, H, L), dtype=torch.int32)
            cluster_bit_counts = torch.zeros((N, H, k, n_buckets),
                                             dtype=torch.int32)
            sequence_lengths = torch.ones((N,), dtype=torch.int32) * L
            sequence_lengths.random_(1, L+1)
                
            s = time.time()
            for i in range(50):
                cluster(
                    hashes, sequence_lengths,
                    groups=groups, counts=counts, centroids=centroids,
                    distances=distances, bitcounts=cluster_bit_counts,
                    iterations=n_iterations,
                    bits=n_buckets
                )
            e = time.time()
            t_clustering = e - s

            print("Clustering with {} bits took {} time".format(n_buckets,
                                                                t_clustering))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import os
import time
import unittest

import torch

from fast_transformers.sparse_product import sparse_dot_product


class TestSparseProductCPU(unittest.TestCase):
    def test_simple_product(self):
        X = torch.randn(10, 4, 100, 32)
        Y = torch.randn(10, 4, 100, 32)
        topk = (torch.cumsum(torch.rand(10, 4, 100, 10)*10, dim=-1)).long()

        products = sparse_dot_product(
            X,
            Y,
            topk,
        )

        all_products = torch.einsum("nhle,nhse->nhls", X, Y)
        self.assertLess(
            torch.max(torch.abs(
                products -
                all_products[
                    torch.arange(10).view(10, 1, 1, 1),
                    torch.arange(4).view(1, 4, 1, 1),
                    torch.arange(100).view(1, 1, 100, 1),
                    topk
                ]
            )),
            1e-4
        )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_small_benchmark(self):
        N = 12
        H = 8
        L = 1000
        S = 1000
        E = 32
        k = 32
        X = torch.randn(N, H, L, E)
        Y = torch.randn(N, H, S, E)
        topk = (torch.cumsum(torch.rand(N, H, L, k)*40, dim=-1)).long()

        n_runs = 10
        s = time.time()
        for run in range(n_runs):
            products = sparse_dot_product(
                X,
                Y,
                topk,
            )
        e = time.time()
        t_s = (e - s) / n_runs

        s = time.time()
        for run in range(n_runs):
            torch.einsum("nhle,nhse->nhls", X, Y)
        e = time.time()
        t_f = (e - s) / n_runs
        print("Sparse: {}, Full: {}, F/S: {}".format(t_s, t_f, t_f/t_s))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

from collections import defaultdict

from .base import BaseBuilder
from ..attention_registry import \
    AttentionRegistry, \
    RecurrentAttentionRegistry, \
    RecurrentCrossAttentionRegistry


class BaseAttentionBuilder(BaseBuilder):
    def __init__(self, registry):
        self._registry = registry
        self._parameters = defaultdict(lambda: None)

    @property
    def available_attentions(self):
        """Return a list with the available attention implementations."""
        return self._registry.keys

    def validate_attention_type(self, attention_type):
        """Parse the attention type according to the rules used by `get()` and
        check if the requested attention is constructible."""
        return all(
            all(t in self._registry for t in a.split(","))
            for a in attention_type.split(":")
        )

    def __setattr__(self, key, value):
        # Make sure we have normal behaviour for the class members _registry
        # and _parameters
        if key in ["_registry", "_parameters"]:
            return object.__setattr__(self, key, value)

        # Assign everything else in the parameters dictionary
        if not self._registry.contains_parameter(key):
            raise AttributeError(("{!r} is not a valid attention "
                                  "parameter name").format(key))
        self._parameters[key] = self._registry.validate_parameter(key, value)

    def __getattr__(self, key):
        if key in self._parameters:
            return self._parameters[key]
        else:
            raise AttributeError()

    def __repr__(self):
        return (
            "{}.from_kwargs(\n".format(self.__class__.__name__) + 
            "\n".join(["    {}={!r},".format(k, v)
                       for k, v in self._parameters.items()])[:-1] +
            "\n)"
        )

    def get(self, attention_type):
        """Construct the attention implementation object and return it.

        The passed in attention_type argument defines the attention to be
        created. It should be a string and in its simplest form it should
        be one of the available choices from `available_attentions`.

        However, to enable attention decoration, namely an attention
        implementation augmenting the functionality of another implementation,
        the attention type can be a colon separated list of compositions like
        the following examples:

            - 'att1' means instantiate att1
            - 'att2:att1' means instantiate att1 and decorate it with att2
            - 'att3:att1,att4' means instantiate att1 and att4 and decorate
              them with att3

        Arguments
        ---------
            attention_type: A string that contains one or more keys from
                            `available_attentions` separated with a colon to
                            denote the decoration pattern.
        """
        compositions = reversed(attention_type.split(":"))
        attentions = []
        for c in compositions:
            attentions = [
                self._construct_attention(t, attentions)
                for t in c.split(",")
            ]
        if len(attentions) > 1:
            raise ValueError(("Invalid attention_type argument "
                              "{!r}").format(attention_type))
        return attentions[0]
        
    def _construct_attention(self, attention_type, decorated=[]):
        """Construct an attention implementation object.

        Arguments
        ---------
            attention_type: A string that contains a single key from the
                            `available_attentions`
            decorated: A list of attention implementations to pass as arguments
                       to be decorated
        """
        if attention_type not in self._registry:
            raise ValueError(("Unknown attention type "
                              "{!r}").format(attention_type))

        attention, parameters = self._registry[attention_type]
        parameter_dictionary = {
            p: self._registry.validate_parameter(p, self._parameters[p])
            for p in parameters
        }

        return attention(*decorated, **parameter_dictionary)


class AttentionBuilder(BaseAttentionBuilder):
    """Build attention implementations for batch sequence processing or
    training."""
    def __init__(self):
        super(AttentionBuilder, self).__init__(AttentionRegistry)


class RecurrentAttentionBuilder(BaseAttentionBuilder):
    """Build attention implementations for autoregressive sequence
    processing."""
    def __init__(self):
        super(RecurrentAttentionBuilder, self).__init__(
            RecurrentAttentionRegistry
        )


class RecurrentCrossAttentionBuilder(BaseAttentionBuilder):
    """Build attention implementations for autoregressive cross attention
    computation."""
    def __init__(self):
        super(RecurrentCrossAttentionBuilder, self).__init__(
            RecurrentCrossAttentionRegistry
        )



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""Provide a library with fast transformer implementations."""

__author__ = "Angelos Katharopoulos, Apoorv Vyas"
__copyright__ = "Copyright (c) 2020 Idiap Research Institute"
__license__ = "MIT"
__maintainer__ = "Angelos Katharopoulos, Apoorv Vyas"
__email__ = "angelos.katharopoulos@idiap.ch, avyas@idiap.ch"
__url__ = "https://github.com/idiap/fast-transformers"
__version__ = "0.2.1"



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import os
import time
from os import getenv
import unittest

import torch

from fast_transformers.sparse_product import sparse_dot_product


class TestSparseProductBackward(unittest.TestCase):
    @property
    def device(self):
        return "cpu"

    def _zero_grad(self, Q, K):
        for x in [Q, K]:
            if x.grad is not None:
                x.grad[...] = 0

    def test_simple_grad(self):
        N = 2
        H = 4
        L = 100
        S = 100
        E = 32
        k = 10
        Q = torch.randn(N, H, L, E).to(self.device).requires_grad_(True)
        K = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)
        topk = torch.round(
                torch.cumsum(torch.rand(N, H, L, k)*10, dim=-1)
            ).long().to(self.device)

        self._zero_grad(Q, K)
        QK_full = torch.einsum("nhle,nhse->nhls", Q, K)
        QK_selected = QK_full[
            torch.arange(N).view(N, 1, 1, 1).to(self.device),
            torch.arange(H).view(1, H, 1, 1).to(self.device),
            torch.arange(L).view(1, 1, L, 1).to(self.device),
            topk
        ]
        QK_selected.sum().backward()
        grad = [torch.clone(Q.grad), torch.clone(K.grad)]

        self._zero_grad(Q, K)
        QK_selected_hat = sparse_dot_product(Q, K, topk)
        QK_selected_hat.sum().backward()
        grad_hat = [torch.clone(Q.grad), torch.clone(K.grad)]

        self.assertLess(
            torch.abs(QK_selected - QK_selected_hat).max(),
            1e-4
        )
        for g1, g2 in zip(grad, grad_hat):
            self.assertLess(
                torch.abs(g1 - g2).max(),
                1e-4
            )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_benchmark_forward(self):
        N = 12
        H = 8
        L = 1024
        S = 1024
        E = 32
        k = 32
        Q = torch.randn(N, H, L, E).to(self.device).requires_grad_(True)
        K = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)
        topk = torch.round(
                torch.cumsum(torch.rand(N, H, L, k)*(S//k), dim=-1)
            ).long().to(self.device)
        n_runs = 10
        s = time.time()
        for i in range(n_runs):
            QK = torch.einsum("nhle,nhse->nhls", Q, K)
            QK.sum()
        e = time.time()
        t_full = (e - s) / n_runs

        s = time.time()
        for i in range(n_runs):
            QK = sparse_dot_product(Q, K, topk)
            QK.sum()
        e = time.time()
        t_sparse = (e - s) / n_runs
        print("Benchmark Forward: T_Full: {}, T_Sparse: {}".format(t_full, t_sparse))

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_benchmark_forward_backward(self):
        N = 12
        H = 8
        L = 1024
        S = 1024
        E = 32
        k = 32
        Q = torch.randn(N, H, L, E).to(self.device).requires_grad_(True)
        K = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)
        topk = torch.round(
                torch.cumsum(torch.rand(N, H, L, k)*(S//k), dim=-1)
            ).long().to(self.device)
        n_runs = 10
        self._zero_grad(Q, K)
        s = time.time()
        for i in range(n_runs):
            QK = torch.einsum("nhle,nhse->nhls", Q, K)
            QK.sum().backward()
        e = time.time()
        t_full = (e - s) / n_runs

        self._zero_grad(Q, K)
        s = time.time()
        for i in range(n_runs):
            QK = sparse_dot_product(Q, K, topk)
            QK.sum().backward()
        e = time.time()
        t_sparse = (e - s) / n_runs
        print("Benchmark Forward-Backward: T_Full: {}, T_Sparse: {}".format(t_full, t_sparse))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Similar to the corresponding module in fast_transformers.attention, this
module performs all the query, key, value projections and output projections
leaving the implementation of the attention to the inner attention module.

The crucial difference with respect to the self attention recurrent module
(fast_transformers.recurrent.attention.RecurrentAttentionLayer) is that it
doesn't recompute the projections for the keys and values if the state is not
None.
"""

from torch.nn import Linear, Module


class RecurrentCrossAttentionLayer(Module):
    """See fast_transformers.attention.attention_layer.AttentionLayer .

    The differences with the aforementioned module as well as the
    RecurrentAttentionLayer are that this module projects the query every time
    and the keys and values only the first time they are provided.

    Arguments
    ---------
        attention: Specific inner attention implementation that just computes a
                   weighted average of values given a similarity of queries and
                   keys.
        d_model: The input feature dimensionality
        n_heads: The number of heads for the multi head attention
        d_keys: The dimensionality of the keys/queries
                (default: d_model/n_heads)
        d_values: The dimensionality of the values (default: d_model/n_heads)
    """
    def __init__(self, attention, d_model, n_heads, d_keys=None,
                 d_values=None):
        super(RecurrentCrossAttentionLayer, self).__init__()

        # Fill d_keys and d_values
        d_keys = d_keys or (d_model//n_heads)
        d_values = d_values or (d_model//n_heads)

        self.inner_attention = attention
        self.query_projection = Linear(d_model, d_keys * n_heads)
        self.key_projection = Linear(d_model, d_keys * n_heads)
        self.value_projection = Linear(d_model, d_values * n_heads)
        self.out_projection = Linear(d_values * n_heads, d_model)
        self.n_heads = n_heads

    def forward(self, query, keys, values, key_lengths, state=None):
        """Attend to the keys and values based on the passed in query.

        In the argument description we make use of the following sizes

            - N: the batch size
            - S: the sequence length of the keys and values
            - D: The input feature dimensionality passed in the constructor as
              'd_model'

        Argument
        --------
            query: (N, D) The tensor containing the queries
            keys: (N, S, D) The tensor containing the keys
            values: (N, S, D) The tensor containing the values
            key_lengths: A fast_transformers.masking.BaseMask implementation
                         that defines the length of each key/value sequence
            state: The state varies depending on the inner attention
                   implementation, but if it is not None then the keys and
                   values are ignored
        """
        #Extract some shapes
        N, _ = query.shape
        H = self.n_heads

        # Project the query
        query = self.query_projection(query).view(N, H, -1)

        # Project the keys and values if there is no state
        if state is None:
            _, S, _ = keys.shape
            keys = self.key_projection(keys).view(N, S, H, -1)
            values = self.value_projection(values).view(N, S, H, -1)
        else:
            keys = None
            values = None

        new_value, state = self.inner_attention(
            query,
            keys,
            values,
            key_lengths,
            state=state
        )
        new_value = new_value.view(N, -1)

        # Project the output and return
        return self.out_projection(new_value), state



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import unittest
import os
import time

import numpy as np

import torch

from fast_transformers.hashing import compute_hashes
from fast_transformers.clustering.hamming import cluster

def simple_lsh(X, A):
    B = (torch.einsum("nj,ij->ni", [X, A]) > 0).long()
    bits = 2**torch.arange(A.shape[0])
    return torch.einsum("ni,i->n", [B, bits])


def generate_hash(n_points, d, b):
    x = torch.rand(n_points, d)
    a = torch.randn(b, d+1)
    a[:,d] = 0
    h = torch.zeros(n_points, dtype=torch.int64)
    compute_hashes(x, a, h)
    return h


def hamming_distance(a, b):
    return sum(c1 != c2 for c1, c2 in zip(a, b))


def get_hamming_distances(data, closest_clusters, n_buckets):
    binary_repr_vectorized = np.vectorize(np.binary_repr)
    hamming_distance_vectorized = np.vectorize(hamming_distance)
    b_data = binary_repr_vectorized(data, n_buckets+1)
    b_closest_clusters = binary_repr_vectorized(closest_clusters, n_buckets+1)
    distance = hamming_distance_vectorized(b_data, b_closest_clusters)
    return distance


def labels_to_centroids(labels, centroids):
    res = [list(map(c.__getitem__, l)) for c,l in zip(centroids,labels)]
    return np.asarray(res)


def verify_distances(data, labels, centroids, distances, n_buckets, lengths):
    closest_centroids = labels_to_centroids(labels, centroids)
    distances_np = get_hamming_distances(data, closest_centroids, n_buckets)
    for idx, l in enumerate(lengths):
        if( l < distances_np.shape[1] ):
            distances_np[idx,l:] = -1

    assert(np.all(distances_np==distances))


class TestClusteringGPU(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not torch.cuda.is_available():
            raise unittest.SkipTest("No CUDA capable device detected")

    def test_clustering_convergence(self):
        N=50
        H=4 
        E=32 
        n_iterations=10
    
        for n_buckets in range(1, 10):
            print('Testing convergence for {} bits'.format(n_buckets))
            k = 2**n_buckets

            L=k
            n_points = L * N * H
            hashes = generate_hash(n_points, E, n_buckets).view(N, H, L).cuda()
            lengths = torch.ones((N,), dtype=torch.int32).cuda() * L
            distances = torch.zeros((N, H, L), dtype=torch.int32).cuda()

            cluster(
                hashes, lengths,
                distances=distances,
                clusters=k,
                iterations=n_iterations,
                bits=n_buckets
            )
    
            distances_np = distances.data.cpu().numpy()
            self.assertEqual(distances_np.sum(), 0)

    def test_clustering(self):
        N=50
        H=4
        L=100
        E=32 

        k=20
        n_buckets=31
        n_iterations=10

        n_points = L * N * H

        groups = torch.zeros((N, H, L), dtype=torch.int32).cuda()
        counts = torch.zeros((N, H, k), dtype=torch.int32).cuda()
        centroids = torch.zeros((N, H, k), dtype=torch.int64).cuda()
        distances = torch.zeros((N, H, L), dtype=torch.int32).cuda()
        cluster_bit_counts = torch.zeros((N, H, k, n_buckets),
                                         dtype=torch.int32).cuda()
        sequence_lengths = torch.ones((N,), dtype=torch.int32).cuda() * L

        for i in range(50):
            hashes = generate_hash(n_points, E, n_buckets).view(N, H, L).cuda()
    
            cluster(
                hashes, sequence_lengths,
                groups=groups, counts=counts, centroids=centroids,
                distances=distances, bitcounts=cluster_bit_counts,
                iterations=n_iterations,
                bits=n_buckets
            )
    
            lengths_np = sequence_lengths.repeat_interleave(H).cpu().numpy()
            hashes_np = hashes.view(N * H, L).cpu().numpy()
            groups_np = groups.view(N * H, L).cpu().numpy()
            distances_np = distances.view(N * H, L).cpu().numpy()
            centroids_np = centroids.view(N * H, k).cpu().numpy()
    
            verify_distances(hashes_np, groups_np, centroids_np,
                             distances_np, n_buckets, lengths_np)

    def test_masked_clustering(self):
        N=50
        H=4 
        L=100
        E=32 
        
        k=20
        n_buckets=31
        n_iterations=10

        n_points = L * N * H

        groups = torch.zeros((N, H, L), dtype=torch.int32).cuda()
        counts = torch.zeros((N, H, k), dtype=torch.int32).cuda()
        centroids = torch.zeros((N, H, k), dtype=torch.int64).cuda()
        distances = torch.zeros((N, H, L), dtype=torch.int32).cuda()
        cluster_bit_counts = torch.zeros((N, H, k, n_buckets),
                                         dtype=torch.int32).cuda()
        sequence_lengths = torch.ones((N,), dtype=torch.int32).cuda() * L
    
        for i in range(50):
            hashes = generate_hash(n_points, E, n_buckets).view(N, H, L).cuda()
            sequence_lengths.random_(1, L+1)

            cluster(
                hashes, sequence_lengths,
                groups=groups, counts=counts, centroids=centroids,
                distances=distances, bitcounts=cluster_bit_counts,
                iterations=n_iterations,
                bits=n_buckets
            )
            lengths_np = sequence_lengths.repeat_interleave(H).data.cpu().numpy()
            hashes_np = hashes.view(N * H, L).data.cpu().numpy()
            groups_np = groups.view(N * H, L).data.cpu().numpy()
            distances_np = distances.view(N * H, L).data.cpu().numpy()
            centroids_np = centroids.view(N * H, k).data.cpu().numpy()
            verify_distances(hashes_np, groups_np, centroids_np, 
                             distances_np, n_buckets, lengths_np)

    def test_masked_clustering_convergence(self):
        N=50
        H=4 
        E=32 
        
        n_iterations=10

    
        for n_buckets in range(1, 10):
            print('Testing convergence for {} bits'.format(n_buckets))
            k = 2**n_buckets
            L = k + 1
            n_points = L * N * H

            hashes = generate_hash(n_points, E, n_buckets).view(N, H, L).cuda()

            groups = torch.zeros((N, H, L), dtype=torch.int32).cuda()
            counts = torch.zeros((N, H, k), dtype=torch.int32).cuda()
            centroids = torch.zeros((N, H, k), dtype=torch.int64).cuda()
            distances = torch.zeros((N, H, L), dtype=torch.int32).cuda()
            cluster_bit_counts = torch.zeros((N, H, k, n_buckets),
                                             dtype=torch.int32).cuda()
            sequence_lengths = torch.ones((N,), dtype=torch.int32).cuda() * L
            sequence_lengths.random_(L)
            sequence_lengths += 1

            cluster(
                hashes, sequence_lengths,
                groups=groups, counts=counts, centroids=centroids,
                distances=distances, bitcounts=cluster_bit_counts,
                iterations=n_iterations,
                bits=n_buckets
            )

            lengths_np = sequence_lengths.cpu().numpy()
            distances_np = distances.cpu().numpy()
            for n in range(N):
                distances_np[n, :, lengths_np[n]-1:] = 0
            assert(distances_np.sum() == 0)

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_benchmark_clustering(self):
        N=12
        H=4 
        L=1000
        E=32 
        
        k=100
        n_buckets=63
        n_iterations=10
        
        n_points = L * N * H
        for n_buckets in range(10, 64):        
            hashes = generate_hash(n_points, E, n_buckets).view(N, H, L).cuda()
            groups = torch.zeros((N, H, L), dtype=torch.int32).cuda()
            counts = torch.zeros((N, H, k), dtype=torch.int32).cuda()
            centroids = torch.zeros((N, H, k), dtype=torch.int64).cuda()
            distances = torch.zeros((N, H, L), dtype=torch.int32).cuda()
            cluster_bit_counts = torch.zeros((N, H, k, n_buckets),
                                             dtype=torch.int32).cuda()
            sequence_lengths = torch.ones((N,), dtype=torch.int32).cuda() * L
            sequence_lengths.random_(1, L+1)

            for i in range(500):
                cluster(
                    hashes, sequence_lengths,
                    groups=groups, counts=counts, centroids=centroids,
                    distances=distances, bitcounts=cluster_bit_counts,
                    iterations=n_iterations,
                    bits=n_buckets
                )

            s = torch.cuda.Event(enable_timing=True)
            e = torch.cuda.Event(enable_timing=True)
            s.record()
            cluster(
                hashes, sequence_lengths,
                groups=groups, counts=counts, centroids=centroids,
                distances=distances, bitcounts=cluster_bit_counts,
                iterations=n_iterations,
                bits=n_buckets
            )   
            e.record()
            torch.cuda.synchronize()
            t_clustering = s.elapsed_time(e)

            print("Clustering with {} bits took {} time".format(
                n_buckets,
                t_clustering)
            )


if __name__ == "__main__":
    unittest.main()




#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import unittest
import os
import time

import torch
from torch.nn.init import normal_
from torch.nn import functional as F

from fast_transformers.hashing import compute_hashes
from fast_transformers.clustering.hamming import cluster
from fast_transformers.aggregate import aggregate, broadcast
from fast_transformers.aggregate import clustered_broadcast

def cluster_queries(Q, query_lengths, C, I, B):
    N, H, L, E = Q.shape
    planes = Q.new_empty((B, E+1))
    normal_(planes)
    planes[:, -1] = 0
    hashes = compute_hashes(Q.view(N*H*L, E), planes).view(N, H, L)
    # Cluster the hashes and return the cluster index per query
    groups, counts =  cluster(
        hashes,
        query_lengths,
        clusters=C,
        iterations=I,
        bits=B
    )

    return groups, counts


class TestClusteredBroadcastGPU(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not torch.cuda.is_available():
            raise unittest.SkipTest("No CUDA capable device detected")

    def test_broadcast(self):
        N = 10
        H = 3
        L = 500
        S = 500
        E = 32
        C = 4
        I = 5
        B = 16

        Q = torch.randn(N, H, L, E).cuda()
        lengths = torch.full((N,), L, dtype=torch.int32).cuda()
        lengths[1] = 400
        lengths[3] = 200
        lengths[7] = 450
        lengths[8] = 150
        groups, counts = cluster_queries(Q, lengths, C, I, B)
        Q_grouped = aggregate(Q, groups, 1/counts.float())
        K = torch.randn(N, H, S, E).cuda()
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)

        V = torch.randn(N, H, S, E).cuda()
        A = F.softmax(QK, dim=-1)
        V_new = torch.einsum("nhls,nhse->nhle", A, V)
        V_broadcast = torch.zeros((N, H, L, E), dtype=V_new.dtype).cuda()
        #V_broadcast = broadcast_clustered(V_new, groups, counts, lengths, V_broadcast)
        V_broadcast = clustered_broadcast(V_new, groups, counts, lengths, V_broadcast)

        V_broadcast_2 = broadcast(
            V_new,
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, E), device=Q.device)
        )
        self.assertLess(
            torch.max(torch.abs(
                V_broadcast_2
                - V_broadcast
                )
            ),
            1e-4
        )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_broadcast_benchmark(self):
        N = 12
        H = 8
        L = 1000
        S = 1000
        E = 32
        C = 200
        I = 5
        B = 32

        Q = torch.randn(N, H, L, E).cuda()
        lengths = torch.full((N,), L, dtype=torch.int32).cuda()
        groups, counts = cluster_queries(Q, lengths, C, I, B)
        Q_grouped = aggregate(Q, groups, 1/counts.float())
        K = torch.randn(N, H, S, E).cuda()
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)

        V = torch.randn(N, H, S, E).cuda()
        A = F.softmax(QK, dim=-1)
        V_new = torch.einsum("nhls,nhse->nhle", A, V)
        V_broadcast = torch.zeros((N, H, L, E), dtype=V_new.dtype).cuda()
        #V_broadcast = broadcast_clustered(V_new, groups, counts, lengths, V_broadcast)
        V_broadcast = clustered_broadcast(V_new, groups, counts, lengths, V_broadcast)

        V_broadcast_2 = broadcast(
            V_new,
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, E), device=Q.device)
        )

        self.assertLess(
            torch.max(torch.abs(
                V_broadcast_2
                - V_broadcast
                )
            ),
            1e-4
        )

        for i in range(2000):
            #V_broadcast = broadcast_clustered(V_new, groups, counts, lengths, V_broadcast)
            V_broadcast = clustered_broadcast(V_new, groups, counts, lengths, V_broadcast)


        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        V_broadcast = clustered_broadcast(V_new, groups, counts, lengths, V_broadcast)
        e.record()
        torch.cuda.synchronize()
        t_broadcast = s.elapsed_time(e)

        for i in range(200):
            V_broadcast_2 = broadcast(
                V_new,
                groups,
                torch.ones_like(counts, dtype=torch.float32),
                torch.zeros((N, H, L, E), device=Q.device)
            )


        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        V_broadcast_2 = broadcast(
            V_new,
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, E), device=Q.device)
        )
        e.record()
        torch.cuda.synchronize()
        t_broadcast_2 = s.elapsed_time(e)

        print("B1: {}, B2: {}".format(t_broadcast, t_broadcast_2))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Autoregressive implementations for cross attention as a recurrent module.

The attention implementations in this module expect one input for query and a
sequence of inputs for keys and values. The sequence for the keys and values is
fixed for all queries.

Example
--------

    import torch

    from fast_transformers.recurrent.attention import \
        RecurrentCrossAttentionLayer, RecurrentCrossFullAttention

    att = RecurrentCrossAttentionLayer(RecurrentCrossFullAttention(), 16, 4)
    state = None
    x = torch.rand(8, 16)
    memory = torch.rand(8, 64, 16)
    for i in range(10):
        x, state = att(x, memory, memory, state=state)
"""

from .attention_layer import RecurrentCrossAttentionLayer
from .full_attention import RecurrentCrossFullAttention
from .linear_attention import RecurrentCrossLinearAttention



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import unittest

import numpy as np
import torch

from fast_transformers.causal_product import causal_dot_product


class TestCausalProduct(unittest.TestCase):
    def _zero_grad(self, *tensors):
        for t in tensors:
            if t.grad is not None:
                t.grad[...] = 0

    def _test_api(self, device):
        for t in range(10):
            N = 2
            H = 4
            L = 100
            S = 100
            E = np.random.randint(10, 256)
            M = np.random.randint(10, 256)
            Q = torch.rand(N, H, L, E).to(device).requires_grad_(True)
            K = torch.rand(N, H, S, E).to(device).requires_grad_(True)
            V = torch.randn(N, H, S, M).to(device).requires_grad_(True)

            self._zero_grad(Q, K, V)
            QK = torch.einsum("nhle,nhse->nhls", Q, K)
            mask = torch.tril(torch.ones(L, S))[None, None].to(device)
            QK = QK * mask
            QK = QK / (QK.sum(-1, keepdim=True) + 1e-6)
            V_new = torch.einsum("nhls,nhsm->nhlm", QK, V)
            V_new.sum().backward()
            grad = [torch.clone(x.grad) for x in [Q, K, V]]

            self._zero_grad(Q, K, V)
            V_new_hat = causal_dot_product(Q, K, V)
            Z = torch.einsum(
                "nhle,nhle->nhl",
                Q,
                torch.cumsum(K, dim=-2) + 1e-6
            ).unsqueeze(-1)

            V_new_hat = V_new_hat / Z
            V_new_hat.sum().backward()
            grad_hat = [torch.clone(x.grad) for x in [Q, K, V]]

            self.assertLess(
                torch.abs(V_new - V_new_hat).max(),
                5e-4
            )
            for g1, g2 in zip(grad, grad_hat):
                self.assertLess(
                    torch.abs(g1-g2).max(),
                    5e-4
                )

    def test_api_cpu(self):
        self._test_api("cpu")

    @unittest.skipUnless(torch.cuda.is_available(), "No CUDA capable device")
    def test_api_cuda(self):
        self._test_api("cuda")


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import unittest
import os
import time

import torch

from fast_transformers.aggregate import aggregate_cpu, broadcast_cpu


class TestAggregateCPU(unittest.TestCase):
    def test_aggregate(self):
        N = 1
        H = 1
        L = 40
        E = 2
        C = 4

        x = torch.rand((N, H, L, E))
        g = (torch.arange(L) % C).view(1, 1, L).repeat(N, H, 1).int()
        c = (.1*torch.ones(N, H, C))
        y = torch.zeros(N, H, C, E)

        aggregate_cpu(x, g, c, y)
        for i in range(C):
            self.assertLess(
                torch.abs(
                    x[:, :, i::C, :].mean(2) - y[:, :, i, :]
                ).max().item(),
                1e-6
            )

    def test_broadcast(self):
        N = 10
        H = 3
        L = 40
        E = 32
        C = 4

        y = torch.rand(N, H, C, E)
        g = (torch.arange(L) % C).view(1, 1, L).repeat(N, H, 1).int()
        c = (.1*torch.ones(N, H, C))
        x = torch.rand((N, H, L, E))

        broadcast_cpu(y, g, c, x)
        for i in range(C):
            self.assertTrue(
                torch.all(x[:, :, i::C] == 0.1*y[:, :, i:i+1, :])
            )

    def test_both(self):
        N = 10
        H = 3
        L = 40
        E = 32
        C = 4

        x_start = torch.rand(N, H, L, E)
        x_end = torch.rand(N, H, L, E)
        g = (torch.rand(N, H, L)*C).int()
        c = torch.zeros(N, H, C)
        y = torch.zeros((N, H, C, E))

        # Aggregating ones should give us the counts
        aggregate_cpu(
            torch.ones(N, H, L, 1),
            g,
            torch.ones(N, H, C),
            c
        )
        for i in range(C):
            self.assertTrue(torch.all((g == i).sum(2) == c[:, :, i].long()))

        # Aggregating into averages twice should be a noop
        aggregate_cpu(x_start, g, 1/c, y)
        broadcast_cpu(y, g, torch.ones(N, H, C), x_start)
        y.zero_()
        aggregate_cpu(x_start, g, 1/c, y)
        broadcast_cpu(y, g, torch.ones(N, H, C), x_end)
        self.assertLess(
            torch.abs(x_start-x_end).max().item(),
            1e-6
        )

    def test_aggregate_masked(self):
        N = 10
        H = 3
        L = 40
        E = 32
        C = 4

        x = torch.rand((N, H, L, E))
        g = (torch.arange(L) % C).view(1, 1, L).repeat(N, H, 1).int()
        g[:, :, -4:] = -1
        c = torch.ones(N, H, C)/9.
        y = torch.zeros(N, H, C, E)

        aggregate_cpu(x, g, c, y)
        for i in range(C):
            self.assertLess(
                torch.abs(
                    x[:, :, i::C, :][:, :, :-1, :].mean(2) - y[:, :, i, :]
                ).max().item(),
                1e-6
            )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_aggregate_benchmark(self):
        N = 12
        H = 8
        L = 1000
        S = 1000
        E = 32
        C = 100
        x = torch.rand((N, H, L, E))
        g = (torch.arange(L) % C).view(L, 1, 1).repeat(1, N, H).int()
        c = 0.1*torch.ones(C, N, H)
        y = torch.zeros((C, N, H, E))

        s = time.time()
        for i in range(100):
            aggregate_cpu(x, g, c, y)
        e = time.time()
        t_aggregate = e - s

        print('Aggregate Time: {}'.format(t_aggregate))

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_broadcast_benchmark(self):
        N = 12
        H = 8
        L = 1000
        S = 1000
        E = 32
        C = 100

        y = torch.rand((N, H, C, E))
        g = (torch.arange(L) % C).view(1, 1, L).repeat(N, H, 1).int()
        c = 0.1*torch.ones(N, H, C)
        x = torch.zeros((N, H, L, E))

        s = time.time()
        for i in range(100):
            broadcast_cpu(y, g, c, x)
        e = time.time()
        t_broadcast = e - s

        print('Broadcast Time: {}'.format(t_broadcast))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Implement the typical softmax attention as a recurrent cross attention
module to speed up autoregressive decoding."""

from math import sqrt

import torch
from torch.nn import Dropout, Module

from ....attention_registry import RecurrentCrossAttentionRegistry, Optional, \
    Float

class RecurrentCrossFullAttention(Module):
    """Implement autoregressive softmax cross attention as a recurrent
    module.

    Arguments
    ---------
        softmax_temp: The temperature to use for the softmax attention.
                      (default: 1/sqrt(d_keys) where d_keys is computed at
                      runtime)
        attention_dropout: The dropout rate to apply to the attention
                           (default: 0.1)
    """
    def __init__(self, softmax_temp=None, attention_dropout=0.1):
        super(RecurrentCrossFullAttention, self).__init__()
        self.softmax_temp = softmax_temp
        self.dropout = Dropout(attention_dropout)

    def forward(self, query, keys, values, key_lengths, state=None):
        # Extract some shapes and compute the temperature
        N, H, E = query.shape
        softmax_temp = self.softmax_temp or 1./sqrt(E)

        # Extract the keys and values either from the arguments or the state
        if state is not None:
            keys, values = state

        # Compute the unnormalized attention and apply the key length mask
        QK = torch.einsum("nhe,nshe->nsh", query, keys)
        QK = QK + key_lengths.additive_matrix[:, :, None]

        # Compute the attention and the weighted average
        A = self.dropout(torch.softmax(softmax_temp * QK, dim=1))
        V = torch.einsum("nsh,nshd->nhd", A, values)

        # Make sure that we return a contiguous value
        return V.contiguous(), [keys, values]


# Register the attention implementation so that it becomes available in our
# builders
RecurrentCrossAttentionRegistry.register(
    "full", RecurrentCrossFullAttention,
    [
        ("softmax_temp", Optional(Float)),
        ("attention_dropout", Optional(Float, 0.1))
    ]
)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#


import unittest

import torch

from fast_transformers.recurrent.attention import RecurrentCrossAttentionLayer


class TestRecurrentCrossAttentionLayer(unittest.TestCase):
    def _assert_sizes_attention(self, qshape, kshape, vshape):
        def inner(q, k, v, kl, state=None):
            if state is not None:
                k, v = state
            self.assertEqual(q.shape, qshape)
            self.assertEqual(k.shape, kshape)
            self.assertEqual(v.shape, vshape)
            N, H, E = q.shape
            _, _, _, D = v.shape
            return v.new_zeros((N, H, D)), [k, v]
        return inner

    def test_forward(self):
        att = RecurrentCrossAttentionLayer(
            self._assert_sizes_attention(
                (10, 4, 25),
                (10, 42, 4, 25),
                (10, 42, 4, 25)
            ),
            100,
            4
        )

        v, s = att(
            torch.rand(10, 100),
            torch.rand(10, 42, 100),
            torch.rand(10, 42, 100),
            None,
            state=None
        )
        self.assertEqual(v.shape, (10, 100))
        self.assertEqual(s[0].shape, (10, 42, 4, 25))
        self.assertEqual(s[1].shape, (10, 42, 4, 25))

        v, s = att(
            torch.rand(10, 100),
            None,
            None,
            None,
            state=[torch.rand(10, 42, 4, 25), torch.rand(10, 42, 4, 25)]
        )
        self.assertEqual(v.shape, (10, 100))
        self.assertEqual(s[0].shape, (10, 42, 4, 25))
        self.assertEqual(s[1].shape, (10, 42, 4, 25))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""Implement transformer encoders and decoders that are going to be used with
different attention mechanisms.

In all cases the batch dimension is first and the sequence dimension is second.
"""

import torch
from torch.nn import Dropout, LayerNorm, Linear, Module, ModuleList
import torch.nn.functional as F

from .masking import FullMask, LengthMask


class TransformerEncoderLayer(Module):
    """Self attention and feed forward network with skip connections.
    
    This transformer encoder layer implements the same encoder layer as
    PyTorch but is a bit more open for extension by receiving the attention
    implementation as a constructor argument.

    Arguments
    ---------
        attention: The attention implementation to use given as a nn.Module
        d_model: The input feature dimensionality
        n_heads: The number of heads for the multi head attention
        d_ff: The dimensionality of the intermediate features after the
              attention (default: d_model*4)
        dropout: The dropout rate to apply to the intermediate features
                 (default: 0.1)
        activation: {'relu', 'gelu'} Which activation to use for the feed
                    forward part of the layer (default: relu)
    """
    def __init__(self, attention, d_model, n_heads, d_ff=None, dropout=0.1,
                 activation="relu"):
        super(TransformerEncoderLayer, self).__init__()
        d_ff = d_ff or 4*d_model
        self.attention = attention
        self.linear1 = Linear(d_model, d_ff)
        self.linear2 = Linear(d_ff, d_model)
        self.norm1 = LayerNorm(d_model)
        self.norm2 = LayerNorm(d_model)
        self.dropout = Dropout(dropout)
        self.activation = F.relu if activation == "relu" else F.gelu

    def forward(self, x, attn_mask=None, length_mask=None):
        """Apply the transformer encoder to the input x.

        Arguments
        ---------
            x: The input features of shape (N, L, E) where N is the batch size,
               L is the sequence length (padded) and E is d_model passed in the
               constructor.
            attn_mask: An implementation of fast_transformers.masking.BaseMask
                       that encodes where each element of x can attend to.
            length_mask: An implementation of
                         fast_transformers.masking.BaseMask that encodes how
                         many elements each sequence in the batch consists of.
        """
        # Normalize the masks
        N = x.shape[0]
        L = x.shape[1]
        attn_mask = attn_mask or FullMask(L, device=x.device)
        length_mask = length_mask or \
            LengthMask(x.new_full((N,), L, dtype=torch.int64))

        # Run self attention and add it to the input
        x = x + self.dropout(self.attention(
            x, x, x,
            attn_mask=attn_mask,
            query_lengths=length_mask,
            key_lengths=length_mask
        ))

        # Run the fully connected part of the layer
        y = x = self.norm1(x)
        y = self.dropout(self.activation(self.linear1(y)))
        y = self.dropout(self.linear2(y))

        return self.norm2(x+y)


class TransformerEncoder(Module):
    """TransformerEncoder is little more than a sequence of transformer encoder
    layers.

    It contains an optional final normalization layer as well as the ability to
    create the masks once and save some computation.

    Arguments
    ---------
        layers: list, TransformerEncoderLayer instances or instances that
                implement the same interface.
        norm_layer: A normalization layer to be applied to the final output
                    (default: None which means no normalization)
    """
    def __init__(self, layers, norm_layer=None):
        super(TransformerEncoder, self).__init__()
        self.layers = ModuleList(layers)
        self.norm = norm_layer

    def forward(self, x, attn_mask=None, length_mask=None):
        """Apply all transformer encoder layers to the input x.

        Arguments
        ---------
            x: The input features of shape (N, L, E) where N is the batch size,
               L is the sequence length (padded) and E is d_model passed in the
               constructor of each transformer encoder layer.
            attn_mask: An implementation of fast_transformers.masking.BaseMask
                       that encodes where each element of x can attend to.
            length_mask: An implementation of
                         fast_transformers.masking.BaseMask that encodes how
                         many elements each sequence in the batch consists of.
            
        """
        # Normalize the masks
        N = x.shape[0]
        L = x.shape[1]
        attn_mask = attn_mask or FullMask(L, device=x.device)
        length_mask = length_mask or \
            LengthMask(x.new_full((N,), L, dtype=torch.int64))

        # Apply all the transformers
        for layer in self.layers:
            x = layer(x, attn_mask=attn_mask, length_mask=length_mask)

        # Apply the normalization if needed
        if self.norm is not None:
            x = self.norm(x)

        return x


class TransformerDecoderLayer(Module):
    """The decoder layer from "Attention Is All You Need".

    Similar to the encoder layer, this layer implements the decoder that
    PyTorch implements but can be used with any attention implementation
    because it receives the attention layers as constructor arguments.

    Arguments
    ---------
        self_attention: The attention implementation to use for self attention
                        given as a nn.Module
        cross_attention: The attention implementation to use for cross
                         attention given as a nn.Module
        d_model: The input feature dimensionality
        d_ff: The dimensionality of the intermediate features after the
              attention (default: d_model*4)
        dropout: The dropout rate to apply to the intermediate features
                 (default: 0.1)
        activation: {'relu', 'gelu'} Which activation to use for the feed
                    forward part of the layer (default: relu)
    """
    def __init__(self, self_attention, cross_attention, d_model, d_ff=None,
                 dropout=0.1, activation="relu"):
        super(TransformerDecoderLayer, self).__init__()
        d_ff = d_ff or 4*d_model
        self.self_attention = self_attention
        self.cross_attention = cross_attention
        self.linear1 = Linear(d_model, d_ff)
        self.linear2 = Linear(d_ff, d_model)
        self.norm1 = LayerNorm(d_model)
        self.norm2 = LayerNorm(d_model)
        self.norm3 = LayerNorm(d_model)
        self.dropout = Dropout(dropout)
        self.activation = F.relu if activation == "relu" else F.gelu

    def forward(self, x, memory, x_mask=None, x_length_mask=None,
                memory_mask=None, memory_length_mask=None):
        """Apply the transformer decoder to the input x using the memory
        `memory`.

        Arguments
        ---------
            x: The input features of shape (N, L, E) where N is the batch size,
               L is the sequence length (padded) and E should be the same as
               the d_model passed in the constructor.
            memory: The memory features of shape (N, L', E) where N is the
                    batch size, L' is the memory's sequence length (padded) and
                    E should be the same as the d_model.
            x_mask: An implementation of fast_transformers.masking.BaseMask
                    that encodes where each element of x can attend to in x.
                    Namely the self attention mask.
            x_length_mask: An implementation of a BaseMask that encodes how
                           many elements each sequence in the batch consists
                           of.
            memory_mask: An implementation of BaseMask that encodes where each
                         element of x can attend to in the memory. Namely the
                         cross attention mask.
            memory_length_mask: An implementation of a BaseMask that encodes how
                                many elements each memory sequence in the batch
                                consists of.
        """
        # Normalize the masks
        N = x.shape[0]
        L = x.shape[1]
        L_prime = memory.shape[1]
        x_mask = x_mask or FullMask(L, device=x.device)
        x_length_mask = x_length_mask  or \
            LengthMask(x.new_full((N,), L, dtype=torch.int64))
        memory_mask = memory_mask or FullMask(L, L_prime, device=x.device)
        memory_length_mask = memory_length_mask or \
            LengthMask(x.new_full((N,), L_prime, dtype=torch.int64))

        # First apply the self attention and add it to the input
        x = x + self.dropout(self.self_attention(
            x, x, x,
            attn_mask=x_mask,
            query_lengths=x_length_mask,
            key_lengths=x_length_mask
        ))
        x = self.norm1(x)

        # Secondly apply the cross attention and add it to the previous output
        x = x + self.dropout(self.cross_attention(
            x, memory, memory,
            attn_mask=memory_mask,
            query_lengths=x_length_mask,
            key_lengths=memory_length_mask
        ))

        # Finally run the fully connected part of the layer
        y = x = self.norm2(x)
        y = self.dropout(self.activation(self.linear1(y)))
        y = self.dropout(self.linear2(y))

        return self.norm3(x+y)


class TransformerDecoder(Module):
    """TransformerDecoder is little more than a sequence of transformer decoder
    layers.

    It contains an optional final normalization layer as well as the ability to
    create the masks once and save some computation.

    Arguments
    ----------
        layers: list, TransformerDecoderLayer instances or instances that
                implement the same interface
        norm_layer: A normalization layer to be applied to the final output
                    (default: None which means no normalization)
    """
    def __init__(self, layers, norm_layer=None):
        super(TransformerDecoder, self).__init__()
        self.layers = ModuleList(layers)
        self.norm = norm_layer

    def forward(self, x, memory, x_mask=None, x_length_mask=None,
                memory_mask=None, memory_length_mask=None):
        # Normalize the masks
        N = x.shape[0]
        L = x.shape[1]
        L_prime = memory.shape[1]
        x_mask = x_mask or FullMask(L, device=x.device)
        x_length_mask = x_length_mask  or \
            LengthMask(x.new_full((N,), L, dtype=torch.int64))
        memory_mask = memory_mask or FullMask(L, L_prime, device=x.device)
        memory_length_mask = memory_length_mask or \
            LengthMask(x.new_full((N,), L_prime, dtype=torch.int64))

        # Apply all the transformer decoders
        for layer in self.layers:
            x = layer(x, memory, x_mask=x_mask, x_length_mask=x_length_mask,
                      memory_mask=memory_mask,
                      memory_length_mask=memory_length_mask)

        # Apply the normalization if needed
        if self.norm is not None:
            x = self.norm(x)

        return x



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

import warnings


def check_state(state=None, memory=None):
    if memory is not None:
        warnings.warn(("'memory' is deprecated for recurrent transformers "
                       " and will be removed in the future, use 'state' "
                       "instead"), DeprecationWarning)
    if state is None:
        state = memory
    return state



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import os
import time
import unittest

import torch

from fast_transformers.attention import FullAttention
from fast_transformers.masking import TriangularCausalMask, LengthMask
from fast_transformers.recurrent.attention import RecurrentFullAttention


class TestRecurrentFullAttention(unittest.TestCase):
    def test_forward(self):
        # Prepare the inputs
        N = 10
        H = 4
        E = 25
        M = 64
        L = 100
        q = torch.rand(N, H, E)
        k = torch.rand(N, H, E)
        v = torch.rand(N, H, M)
        memory = [
            torch.rand(N, H, L, E),
            torch.rand(N, H, L, M)
        ]

        # Test the attention module
        att = RecurrentFullAttention(softmax_temp=1)
        v_new, mem_new = att(q, k, v)
        self.assertEqual(v_new.shape, (N, H, M))
        self.assertEqual(len(mem_new), 2)
        self.assertEqual(mem_new[0].shape, (N, H, 1, E))
        self.assertEqual(mem_new[1].shape, (N, H, 1, M))
        v_new, mem_new = att(q, k, v, mem_new)
        self.assertEqual(v_new.shape, (N, H, M))
        self.assertEqual(len(mem_new), 2)
        self.assertEqual(mem_new[0].shape, (N, H, 2, E))
        self.assertEqual(mem_new[1].shape, (N, H, 2, M))

        v_new, mem_new = att(q, k, v, memory)
        self.assertEqual(v_new.shape, (N, H, M))
        self.assertEqual(len(mem_new), 2)
        self.assertEqual(mem_new[0].shape, (N, H, L+1, E))
        self.assertEqual(mem_new[1].shape, (N, H, L+1, M))

    def test_correctness(self):
        # Prepare the inputs
        N = 10
        H = 4
        E = 25
        M = 64
        L = 100
        q = torch.rand(N, L, H, E)
        k = torch.rand(N, L, H, E)
        v = torch.rand(N, L, H, M)
        m1 = TriangularCausalMask(L)
        m2 = LengthMask(torch.full((N,), L, dtype=torch.int64))
        m3 = LengthMask(torch.full((N,), L, dtype=torch.int64))
        att = FullAttention()
        rec_att = RecurrentFullAttention()
        att.eval()
        rec_att.eval()

        v1 = att(q, k, v, m1, m2, m3)
        v2 = []
        memory = None
        for i in range(L):
            v2i, memory = rec_att(q[:, i], k[:, i], v[:, i], memory)
            v2.append(v2i)
        v2 = torch.stack(v2, dim=1)
        self.assertLess(torch.abs(v1-v2).max(), 1e-5)

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_benchmark_cpu(self):
        # Prepare the inputs
        N = 10
        H = 12
        E = 25
        M = 64
        L = 100
        q = torch.rand(N, H, E)
        k = torch.rand(N, H, E)
        v = torch.rand(N, H, M)
        memory = None
        att = RecurrentFullAttention(softmax_temp=1)

        start = time.time()
        for i in range(100):
            v, memory = att(q, k, v, memory)
        end = time.time()
        print("CPU Time taken:", (end-start)*1000, "(ms)")

    @unittest.skipUnless(torch.cuda.is_available(), "no CUDA capable device")
    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_benchmark_gpu(self):
        # Prepare the inputs
        N = 10
        H = 12
        E = 25
        M = 64
        L = 100
        q = torch.rand(N, H, E).cuda()
        k = torch.rand(N, H, E).cuda()
        v = torch.rand(N, H, M).cuda()
        memory = None
        att = RecurrentFullAttention(softmax_temp=1)

        start = torch.cuda.Event(enable_timing=True)
        end = torch.cuda.Event(enable_timing=True)
        start.record()
        for i in range(100):
            v, memory = att(q, k, v, memory)
        end.record()
        torch.cuda.synchronize()
        print("GPU time taken:", start.elapsed_time(end), "(ms)")


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import os
import time
import unittest

import numpy as np
import torch

from fast_transformers.causal_product.causal_product_cpu import \
    causal_dot_product, causal_dot_backward


def max_relative_error(a, b, eps=1e-6):
    return torch.abs((a-b) / (torch.abs(a) + eps)).max().item()


class TestCausalProductCPU(unittest.TestCase):
    kernels = {
        "normal": {
            "forward": causal_dot_product,
            "backward": causal_dot_backward
        },
        # map the optimized versions to other keys here
    }

    def _test_result_forward(self, CP):
        for t in range(10):
            N = 10
            L = 100
            H = 10
            E = np.random.randint(10, 256)
            M = np.random.randint(10, 256)
            Q = torch.rand(N, H, L, E)
            K = torch.rand(N, H, L, E)
            V = torch.rand(N, H, L, M)
            out = torch.zeros(N, H, L, M)
            self.kernels[CP]["forward"](Q, K, V, out)

            QK = torch.einsum("nhle,nhse->nhls", Q, K)
            mask = torch.tril(torch.ones(L, L))
            out2 = torch.einsum("nhls,nhsm,ls->nhlm", QK, V, mask)

            self.assertLess(max_relative_error(out2, out), 1e-5)

    def _test_result_backward(self, CP):
        for t in range(10):
            N = 10
            L = 100
            H = 10
            E = np.random.randint(10, 256)
            M = np.random.randint(10, 256)
            Q = torch.rand(N, H, L, E)
            K = torch.rand(N, H, L, E)
            V = torch.rand(N, H, L, M)
            go = torch.rand(N, H, L, M)
            gq, gk, gv = [torch.zeros_like(x) for x in [Q, K, V]]
            self.kernels[CP]["backward"](Q, K, V, go, gq, gk, gv)

            Q, K, V = [x.requires_grad_(True) for x in [Q, K, V]]
            QK = torch.einsum("nhle,nhse->nhls", Q, K)
            mask = torch.tril(torch.ones(L, L))
            out = torch.einsum("nhls,nhsm,ls->nhlm", QK, V, mask)
            l = torch.einsum("nhlm,nhlm->", out, go)
            l.backward()

            self.assertLess(max_relative_error(Q.grad, gq), 1e-5)
            self.assertLess(max_relative_error(K.grad, gk), 1e-5)
            self.assertLess(max_relative_error(V.grad, gv), 1e-5)

    def _test_benchmark_forward(self, CP):
        N = 10
        L = 1000
        H = 10
        E = 32
        M = 64
        Q = torch.rand(N, H, L, E)
        K = torch.rand(N, H, L, E)
        V = torch.rand(N, H, L, M)
        out = torch.rand(N, H, L, M)

        # warmup the cache
        for i in range(10):
            self.kernels[CP]["forward"](Q, K, V, out)

        # measure
        start = time.time()
        for i in range(10):
            self.kernels[CP]["forward"](Q, K, V, out)
        end = time.time()
        print("[{}] CPU time taken: {} (ms)".format(
            CP,
            (end-start)*1000
        ))

    def _test_benchmark_backward(self, CP):
        N = 10
        L = 1000
        H = 10
        E = 32
        M = 64
        Q = torch.rand(N, H, L, E)
        K = torch.rand(N, H, L, E)
        V = torch.rand(N, H, L, M)
        go = torch.rand(N, H, L, M)
        gq, gk, gv = [torch.zeros_like(x) for x in [Q, K, V]]

        # warmup the cache
        for i in range(10):
            self.kernels[CP]["backward"](Q, K, V, go, gq, gk, gv)

        # measure
        start = time.time()
        for i in range(10):
            self.kernels[CP]["backward"](Q, K, V, go, gq, gk, gv)
        end = time.time()
        print("[{}] CPU time taken: {} (ms)".format(
            CP,
            (end-start)*1000
        ))

    def test_result_forward(self):
        for k in self.kernels.keys():
            with self.subTest(msg=k):
                self._test_result_forward(k)

    def test_result_backward(self):
        for k in self.kernels.keys():
            with self.subTest(msg=k):
                self._test_result_backward(k)

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_benchmark_forward(self):
        for k in self.kernels.keys():
            with self.subTest(msg=k):
                self._test_benchmark_forward(k)

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_benchmark_backward(self):
        for k in self.kernels.keys():
            with self.subTest(msg=k):
                self._test_benchmark_backward(k)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""Implement clustered self attention."""

from math import sqrt

import torch
import torch.autograd
from torch.nn import Dropout, Module
from torch.nn.init import normal_

from ..attention_registry import AttentionRegistry, Optional, Float, Int, Bool
from ..masking import FullMask
from ..aggregate import aggregate, broadcast
from ..clustering.hamming import cluster
from ..hashing import compute_hashes


class _GroupQueries(torch.autograd.Function):
    @staticmethod
    def forward(ctx, Q, clusters, counts):
        factors = 1/counts.float()
        q_grouped = aggregate(Q, clusters, factors)
        ctx.save_for_backward(clusters, factors)

        return q_grouped

    @staticmethod
    def backward(ctx, grad_q_grouped):
        clusters, factors = ctx.saved_tensors
        grad_q = broadcast(grad_q_grouped, clusters, factors)

        return grad_q, None, None


class _BroadcastValues(torch.autograd.Function):
    @staticmethod
    def forward(ctx, v_grouped, clusters, counts):
        factors = torch.ones_like(counts, dtype=v_grouped.dtype)
        V = broadcast(v_grouped, clusters, factors)
        ctx.save_for_backward(clusters, factors)

        return V

    @staticmethod
    def backward(ctx, grad_v):
        clusters, factors = ctx.saved_tensors
        grad_v_grouped = aggregate(grad_v, clusters, factors)

        return grad_v_grouped, None, None


class ClusteredAttention(Module):
    """Use LSH and clustering in the resulting Hamming space to group queries
    that will have minimal L2 distance from each other.

    Given the queries, keys, and values as Q, K, and V respectively, we
    first cluster the queries in "C" groups and compute the "C" query centroids
    Q_c.

    We now use to the centroids Q_c to compute the attention using:
    
        V'_c = softmax(Q_c.mm(K.t()), dim=-1).mm(V).

    Now the computed values V'_c are "broadcasted" back to the query members
    of the corresponding cluster.

    Arguments
    ---------
        clusters: How many clusters to group the queries into
        iterations: The number of lloyd iterations to perform (default: 10)
        bits: How many bits to use for the hash (default: 32)
        hash_bias: If true, hamming distance proportional to L2 distance
                   If false, hamming distance proportional to cosine distance
                   (default: True)
        softmax_temp: The temperature to use for the softmax attention.
                      (default: 1/sqrt(d_keys) where d_keys is computed at
                      runtime)
        attention_dropout: The dropout rate to apply to the attention
                           (default: 0.1)
    """
    def __init__(self, clusters, iterations=10, bits=32,
                 hash_bias=True, softmax_temp=None, attention_dropout=0.1):
        super(ClusteredAttention, self).__init__()
        self.clusters = clusters
        self.iterations = iterations
        self.bits = bits
        self.hash_bias = hash_bias
        self.softmax_temp = softmax_temp
        self.dropout = Dropout(attention_dropout)

    def _create_query_groups(self, Q, query_lengths):
        N, H, L, E = Q.shape

        # Compute the hashes for all the queries
        planes = Q.new_empty((self.bits, E+1))
        normal_(planes)
        if not self.hash_bias:
            planes[:, -1] = 0
        hashes = compute_hashes(Q.view(N*H*L, E), planes).view(N, H, L)

        # Cluster the hashes and return the cluster index per query
        groups =  cluster(
            hashes,
            query_lengths._lengths.int(),
            clusters=self.clusters,
            iterations=self.iterations,
            bits=self.bits
        )
        return groups

    def _group_queries(self, Q, groups):
        """Aggregate the Qs based on the index of cluster they belong to. Make
        sure to allow for gradient propagation backwards from the grouped
        queries to each query."""
        q_grouped = _GroupQueries.apply(Q, *groups)
        return q_grouped

    def _broadcast_values(self, V, groups):
        """Broadcast the values back to the correct positions but make sure
        that the gradient flows properly."""
        V_new = _BroadcastValues.apply(V.contiguous(), *groups)
        V_new = V_new.permute(0, 2, 1, 3).contiguous()
        return V_new

    def forward(self, queries, keys, values, attn_mask, query_lengths,
                key_lengths):
        # Make sure that there is no attention mask
        assert attn_mask.all_ones, ("Clustered attention cannot use an "
                                    "arbitrary attention mask.")

        queries = queries.permute(0,2,1,3).contiguous()
        keys = keys.permute(0,2,1,3).contiguous()
        values = values.permute(0,2,1,3).contiguous()

        N, H, L, E = queries.shape
        softmax_temp = self.softmax_temp or 1./sqrt(E)
        
        # Cluster the queries into groups
        groups = self._create_query_groups(queries, query_lengths)
        Q_grouped = self._group_queries(queries, groups)
        # Compute the attention
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, keys)
        QK = QK + key_lengths.additive_matrix[:, None, None, :]
        A = self.dropout(torch.softmax(softmax_temp * QK, dim=-1))
        V = torch.einsum("nhls,nhsd->nhld", A, values)

        # Broadcast grouped attention
        return self._broadcast_values(V, groups)


# Register the attention implementation so that it becomes available in our
# builders
AttentionRegistry.register(
    "clustered", ClusteredAttention,
    [
        ("clusters", Int),
        ("iterations", Optional(Int, 10)),
        ("bits", Optional(Int, 32)),
        ("hash_bias", Optional(Bool, True)),
        ("softmax_temp", Optional(Float)),
        ("attention_dropout", Optional(Float, 0.1))
    ]
)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import os
from os import getenv
import time
import unittest

import torch
from torch.nn.init import normal_

from fast_transformers.aggregate import aggregate, broadcast
from fast_transformers.hashing import compute_hashes
from fast_transformers.clustering.hamming import cluster
from fast_transformers.sparse_product import clustered_sparse_weighted_average


def cluster_queries(Q, query_lengths, C, I, B):
    N, H, L, E = Q.shape
    planes = Q.new_empty((B, E+1))
    normal_(planes)
    planes[:, -1] = 0
    hashes = compute_hashes(Q.view(N*H*L, E), planes).view(N, H, L)
    # Cluster the hashes and return the cluster index per query
    groups, counts =  cluster(
        hashes,
        query_lengths,
        clusters=C,
        iterations=I,
        bits=B
    )

    return groups, counts


class TestSparseWeightedAverage(unittest.TestCase):
    @property
    def device(self):
        return "cpu"

    def _zero_grad(self, Q, K):
        for x in [Q, K]:
            if x.grad is not None:
                x.grad[...] = 0

    def test_correctness(self):
        N = 2
        H = 4
        L = 3000
        S = 3000
        E = 32
        k = 32
        C = 100
        I = 10
        B = 32

        Q = torch.randn(N, H, L, E).to(self.device)
        K = torch.randn(N, H, S, E).to(self.device)
        lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
        groups, counts = cluster_queries(Q, lengths, C, I, B)
        Q_grouped = aggregate(Q, groups, 1/counts.float())
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
        _, topk = torch.topk(QK, k, dim=-1)
        topk = topk.contiguous()
        topk_broadcast = broadcast(
            topk.float(),
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, k), device=Q.device)
        )

        weights = torch.rand(N, H, L, k).to(self.device).requires_grad_(True)
        values = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)
        self._zero_grad(weights, values)
        values_selected = values[
            torch.arange(N).view(N, 1, 1, 1).to(self.device),
            torch.arange(H).view(1, H, 1, 1).to(self.device),
            topk_broadcast.long()
        ]
        output = (weights.unsqueeze(-1)*values_selected).sum(-2)
        output.sum().backward()
        grad = [torch.clone(weights.grad), torch.clone(values.grad)]

        self._zero_grad(weights, values)
        output_hat = clustered_sparse_weighted_average(weights, values, topk, groups)
        output_hat.sum().backward()
        grad_hat = [torch.clone(weights.grad), torch.clone(values.grad)]
        self.assertLess(
            torch.abs(output - output_hat).max(),
            1e-4
        )
        for g1, g2 in zip(grad, grad_hat):
            self.assertLess(
                torch.abs(g1 - g2).max(),
                1e-3
            )

    def test_forward(self):
        N = 5
        H = 2 
        L = 100
        S = 100
        E = 32
        C = 10
        I = 10
        B = 32
        k = 5

        Q = torch.randn(N, H, L, E).to(self.device)
        K = torch.randn(N, H, S, E).to(self.device)
        lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
        groups, counts = cluster_queries(Q, lengths, C, I, B)
        Q_grouped = aggregate(Q, groups, 1/counts.float())
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
        _, topk = torch.topk(QK, k, dim=-1)
        topk = topk.contiguous()
        topk_broadcast = broadcast(
            topk.float(),
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, k), device=Q.device)
        )

        weights = torch.rand(N, H, L, k).to(self.device).requires_grad_(True)
        values = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)
        values_selected = values[
            torch.arange(N).view(N, 1, 1, 1).to(self.device),
            torch.arange(H).view(1, H, 1, 1).to(self.device),
            topk_broadcast.long()
        ]

        output = (weights.unsqueeze(-1)*values_selected).sum(-2)
        output_hat = clustered_sparse_weighted_average(weights, values, topk, groups)
        self.assertLess(
            torch.abs(output - output_hat).max(),
            1e-4
        )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_small_forward(self):
        N = 12
        H = 8
        L = 2000
        S = 2000
        E = 32
        k = 32
        C = 100
        I = 10
        B = 32

        Q = torch.randn(N, H, L, E).to(self.device)
        K = torch.randn(N, H, S, E).to(self.device)
        lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
        groups, counts = cluster_queries(Q, lengths, C, I, B)
        Q_grouped = aggregate(Q, groups, 1/counts.float())
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
        _, topk = torch.topk(QK, k, dim=-1)
        topk = topk.contiguous()
        topk_broadcast = broadcast(
            topk.float(),
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, k), device=Q.device)
        )

        weights = torch.rand(N, H, L, k).to(self.device).requires_grad_(True)
        values = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)
       
        n_runs = 20
        s = time.time()
        for i in range(n_runs):
            output_hat = clustered_sparse_weighted_average(
                weights, values,
                topk, groups
            )
        e = time.time()
        t_sparse = (e - s) / n_runs
        print('T_sparse Forward:{}'.format(t_sparse))

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_small_forward_backward(self):
        N = 12
        H = 8
        L = 2000
        S = 2000
        E = 32
        k = 32
        C = 100
        I = 10
        B = 32

        Q = torch.randn(N, H, L, E).to(self.device)
        K = torch.randn(N, H, S, E).to(self.device)
        lengths = torch.full((N,), L, dtype=torch.int32).to(self.device)
        groups, counts = cluster_queries(Q, lengths, C, I, B)
        Q_grouped = aggregate(Q, groups, 1/counts.float())
        QK = torch.einsum("nhle,nhse->nhls", Q_grouped, K)
        _, topk = torch.topk(QK, k, dim=-1)
        topk = topk.contiguous()
        topk_broadcast = broadcast(
            topk.float(),
            groups,
            torch.ones_like(counts, dtype=torch.float32),
            torch.zeros((N, H, L, k), device=Q.device)
        )
        weights = torch.rand(N, H, L, k).to(self.device).requires_grad_(True)
        values = torch.randn(N, H, S, E).to(self.device).requires_grad_(True)

        self._zero_grad(weights, values)
        n_runs = 20
        s = time.time()
        for i in range(n_runs):
            output_hat = clustered_sparse_weighted_average(
                weights, values,
                topk, groups
            )
            output_hat.sum().backward()
        e = time.time()
        t_sparse = (e - s) / n_runs
        print('T_sparse Forward Backward:{}'.format(t_sparse))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""Implement the causally masked linear attention as a recurrent model."""

import torch
from torch.nn import Module

from ....attention_registry import RecurrentAttentionRegistry, Optional, \
    Callable
from ..._utils import check_state


def elu_feature_map(x):
    return torch.nn.functional.elu(x) + 1


class RecurrentLinearAttention(Module):
    """Implement fast_transformers.attention.causal_linear_attention as a
    fixed-dimensional state recurrent model.

    See fast_transformers.attention.linear_attention and
    fast_transformers.attention.causal_linear_attention for the general concept
    of replacing the softmax with feature maps.

    Arguments
    ---------
        feature_map: callable, a callable that applies the feature map to the
                     last dimension of a tensor (default: elu(x)+1)
        eps: float, a small number to ensure the numerical stability of the
             denominator (default: 1e-6)
    """
    def __init__(self, feature_map=None, eps=1e-6):
        super(RecurrentLinearAttention, self).__init__()
        self.feature_map = feature_map or elu_feature_map
        self.eps = eps

    def forward(self, query, key, value, state=None, memory=None):
        # Normalize state/memory
        state = check_state(state, memory)

        # Apply the feature map to the query and key
        Q = self.feature_map(query)
        K = self.feature_map(key)

        # Extract some shapes
        N, H, D = Q.shape
        _, _, M = value.shape

        # Extract the memory or initialize it
        if state is None:
            Si = query.new_zeros((N, H, D, M))
            Zi = query.new_zeros((N, H, D))
        else:
            Si, Zi = state

        # Ensure the batch size did not change
        if len(Si) != N:
            raise ValueError("The batch size changed during iteration")

        # Update the internal state
        #
        # NOTE: The if clause is added due to GitHub PR #10. Simply using the
        # following two lines does not perform the operation in place which
        # means it is slower for inference.
        if K.grad_fn is not None or value.grad_fn is not None:
            Zi = Zi + K
            Si = Si + torch.einsum("nhd,nhm->nhdm", K, value)
        else:
            Zi += K
            Si += torch.einsum("nhd,nhm->nhdm", K, value)

        # Compute the output
        Z = 1. / (torch.einsum("nhd,nhd->nh", Q, Zi) + self.eps)
        V = torch.einsum("nhd,nhdm,nh->nhm", Q, Si, Z)

        return V, [Si, Zi]


# Register the attention implementation so that it becomes available in our
# builders
RecurrentAttentionRegistry.register(
    "linear", RecurrentLinearAttention,
    [("feature_map", Optional(Callable))]
)
RecurrentAttentionRegistry.register(
    "causal-linear", RecurrentLinearAttention,
    [("feature_map", Optional(Callable))]
)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""Similar to the corresponding module in fast_transformers.attention, this
module performs all the query, key, value projections and output projections
leaving the implementation of the attention to the inner attention module."""

from torch.nn import Linear, Module

from ..._utils import check_state


class RecurrentAttentionLayer(Module):
    """See fast_transformers.attention.attention_layer.AttentionLayer.
    
    The only difference with the corresponding module is that this projects
    only one input and then calls the inner attention with the provided
    previous state.

    Arguments
    ---------
        attention: Specific inner attention implementation that just computes a
                   weighted average of values given a similarity of queries and
                   keys.
        d_model: The input feature dimensionality
        n_heads: The number of heads for the multi head attention
        d_keys: The dimensionality of the keys/queries
                (default: d_model/n_heads)
        d_values: The dimensionality of the values (default: d_model/n_heads)
    """
    def __init__(self, attention, d_model, n_heads, d_keys=None,
                 d_values=None):
        super(RecurrentAttentionLayer, self).__init__()

        # Fill d_keys and d_values
        d_keys = d_keys or (d_model//n_heads)
        d_values = d_values or (d_model//n_heads)

        self.inner_attention = attention
        self.query_projection = Linear(d_model, d_keys * n_heads)
        self.key_projection = Linear(d_model, d_keys * n_heads)
        self.value_projection = Linear(d_model, d_values * n_heads)
        self.out_projection = Linear(d_values * n_heads, d_model)
        self.n_heads = n_heads

    def forward(self, query, key, value, state=None, memory=None):
        """Apply attention to the passed in query/key/value after projecting
        them to multiple heads.

        In the argument description we make use of the following sizes

            - N: the batch size
            - D: The input feature dimensionality passed in the constructor as
              'd_model'

        Arguments
        ---------
            query: (N, D) The tensor containing the queries
            key: (N, D) The tensor containing the keys
            value: (N, D) The tensor containing the values
            state: The state varies depending on the inner attention implementation
            memory: **Deprecated** and replaced by state

        Returns
        -------
            The new value for each query as a tensor of shape (N, D).
        """
        # Normalize the state/memory
        state = check_state(state, memory)

        # Project the queries/keys/values
        query = self.query_projection(query)
        key = self.key_projection(key)
        value = self.value_projection(value)

        # Reshape them into many heads and compute the attention
        N, D = query.shape
        H = self.n_heads
        new_value, state = self.inner_attention(
            query.view(N, H, -1),
            key.view(N, H, -1),
            value.view(N, H, -1),
            state
        )
        new_value = new_value.view(N, -1)

        # Project the output and return
        return self.out_projection(new_value), state



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#

"""Spec instances allow to describe and check the type and value of
parameters."""


class Spec(object):
    """Describe and validate a parameter type.

    Arguments
    ---------
        predicate: A callable that checks if the value is acceptable and
                   returns its canonical value or raises ValueError.
        name: A name to create a human readable description of the Spec
    """
    def __init__(self, predicate, name="CustomSpec"):
        self._predicate = predicate
        self._name = name

    def __repr__(self):
        return self._name

    def check(self, x):
        try:
            self._predicate(x)
            return True
        except ValueError:
            return False

    def get(self, x):
        return self._predicate(x)

    def __eq__(self, y):
        return self is y


class Choice(Spec):
    """A parameter type for a set of options.

    Arguments
    ---------
        choices: A set or list of possible values for this parameter
    """
    def __init__(self, choices):
        self._choices = choices

    def get(self, x):
        if x in self._choices:
            return x
        raise ValueError("{!r} is not in {!r}".format(x, self._choices))

    def __repr__(self):
        return "Choice({!r})".format(self._choices)

    def __eq__(self, x):
        if isinstance(x, Choice):
            return self._choices == x._choices
        return False


class _Callable(Spec):
    def __init__(self):
        super(_Callable, self).__init__(None, "Callable")

    def get(self, x):
        if callable(x):
            return x
        raise ValueError("{!r} is not a callable".format(x))


class Optional(Spec):
    """Represent an optional parameter that can either have a value or it can
    be None.

    Arguments
    ---------
        spec: The spec for the value if it is not None
        default: The returned value in case it is None
    """
    def __init__(self, spec, default=None):
        self._other_spec = spec
        self._default = default

    def __repr__(self):
        return "Optional[{!r}, {!r}]".format(self._other_spec, self._default)

    def get(self, x):
        if x is None:
            return self._default
        return self._other_spec.get(x)

    def __eq__(self, x):
        if isinstance(x, Optional):
            return (
                self._other_spec == x._other_spec and
                self._default == x._default
            )
        return False


Int = Spec(int, "Int")
Float = Spec(float, "Float")
Bool = Spec(bool, "Bool")
Callable = _Callable()



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import numpy as np
import torch
import time

from fast_transformers.hashing import compute_hashes
from fast_transformers.clustering.hamming import cluster

def simple_lsh(X, A):
    B = (torch.einsum("nj,ij->ni", [X, A]) > 0).long()
    bits = 2**torch.arange(A.shape[0])
    return torch.einsum("ni,i->n", [B, bits])


def generate_hash(n_points, d, b, h):
    torch.manual_seed(0)
    x = torch.rand(n_points, d).cuda()
    a = torch.randn(b, d + 1).cuda()
    compute_hashes(x, a, h)
    return h


def time_clustering(L, N, H, E,
                    n_batches, n_attentions,
                    k, n_buckets, n_iterations, verbose):
    n_points = L * N * H 
    hashes = torch.zeros(n_points, dtype=torch.int64).cuda()
    hashes = generate_hash(n_points, E, n_buckets, hashes).view(N, H, L)

    groups = torch.zeros((N, H, L), dtype=torch.int32).cuda()
    counts = torch.zeros((N, H, k), dtype=torch.int32).cuda()
    centroids = torch.zeros((N, H, k), dtype=torch.int64).cuda()
    distances = torch.zeros((N, H, L), dtype=torch.int32).cuda()
    cluster_bit_counts = torch.zeros((N, H, k, n_buckets),
                                     dtype=torch.int32).cuda()
    sequence_lengths = torch.ones((N,), dtype=torch.int32).cuda() * L
     
    start = time.time()
    for batch_idx in range(int(n_batches)):
        for attention_idx in range(int(n_attentions)):
            #hashes = generate_hash(n_points, E, n_buckets, hashes).view(L, N, H)
            cluster(
                hashes, sequence_lengths,
                groups=groups, counts=counts, centroids=centroids,
                distances=distances, bitcounts=cluster_bit_counts,
                iterations=n_iterations,
                bits=n_buckets
            )
    end = time.time()
    duration = end - start
    print("Time Elapsed: {}".format(duration))


if __name__ == "__main__":
    L = 1000
    N = 12
    H = 8
    E = 32

    n_batches = 50000/N
    n_attentions = 3
    
    k = 30
    n_buckets = 31
    n_iterations = 10
    verbose = 0
    
    time_clustering(L, N, H, E,
                    n_batches, n_attentions,
                    k, n_buckets, n_iterations, verbose)



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import torch

from .cluster_cpu import cluster as cluster_cpu
try:
    from .cluster_cuda import cluster as cluster_gpu
except ImportError:
    pass


def cluster(
    hashes,
    lengths,
    groups=None,
    counts=None,
    centroids=None,
    distances=None,
    bitcounts=None,
    clusters=30,
    iterations=10,
    bits=32
):
    """Cluster hashes using a few iterations of K-Means with hamming distance.

    All the tensors default initialized to None are optional buffers to avoid
    memory allocations. distances and bitcounts are only used by the CUDA
    version of this call. clusters will be ignored if centroids is provided.

    Arguments
    ---------
        hashes: A long tensor of shape (N, H, L) containing a hashcode for each
                query.
        lengths: An int tensor of shape (N,) containing the sequence length for
                 each sequence in hashes.
        groups: An int tensor buffer of shape (N, H, L) contaning the cluster
                in which the corresponding hash belongs to.
        counts: An int tensor buffer of shape (N, H, K) containing the number
                of elements in each cluster.
        centroids: A long tensor buffer of shape (N, H, K) containing the
                   centroid for each cluster.
        distances: An int tensor of shape (N, H, L) containing the distance to
                   the closest centroid for each hash.
        bitcounts: An int tensor of shape (N, H, K, bits) containing the number
                   of elements that have 1 for a given bit.
        clusters: The number of clusters to use for each sequence. It is
                  ignored if centroids is not None.
        iterations: How many k-means iterations to perform.
        bits: How many of the least-significant bits in hashes to consider.

    Returns
    -------
        groups and counts as defined above.
    """
    device = hashes.device
    N, H, L = hashes.shape

    # Unfortunately cpu and gpu have different APIs so the entire call must be
    # surrounded by an if-then-else
    if device.type == "cpu":
        if groups is None:
            groups = torch.empty((N, H, L), dtype=torch.int32)
        if centroids is None:
            centroids = torch.empty((N, H, clusters), dtype=torch.int64)
        K = centroids.shape[2]
        if counts is None:
            counts = torch.empty((N, H, K), dtype=torch.int32)

        cluster_cpu(
            hashes, lengths,
            centroids, groups, counts,
            iterations, bits
        )

        return groups, counts

    else:
        if groups is None:
            groups = torch.empty((N, H, L), dtype=torch.int32, device=device)
        if centroids is None:
            centroids = torch.empty((N, H, clusters), dtype=torch.int64,
                                    device=device)
        K = centroids.numel() // N // H
        #K = clusters
        if counts is None:
            counts = torch.empty((N, H, K), dtype=torch.int32, device=device)
        if distances is None:
            distances = torch.empty((N, H, L), dtype=torch.int32,
                                    device=device)
        if bitcounts is None:
            bitcounts = torch.empty((N, H, K, bits), dtype=torch.int32,
                                    device=device)
        groups = groups.view(N, H, L)
        counts = counts.view(N, H, K)
        centroids = centroids.view(N, H, K)
        distances = distances.view(N, H, L)
        bitcounts = bitcounts.view(N, H, K, -1)

        cluster_gpu(
            hashes, lengths,
            centroids, distances, bitcounts, groups, counts,
            iterations, bits
        )

        return groups, counts
        



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#


import torch

from .aggregate_cpu import aggregate as aggregate_cpu, \
    broadcast as broadcast_cpu
try:
    from .aggregate_cuda import aggregate as aggregate_gpu, \
        broadcast as broadcast_gpu
    from .clustered_broadcast_cuda import \
        clustered_broadcast as clustered_broadcast_gpu

except ImportError:
    pass


def aggregate(X, G, F, Y=None):
    device = X.device
    if Y is None:
        Y = torch.zeros(
            F.shape + (X.shape[-1],),
            device=device,
            dtype=X.dtype
        )
    else:
        Y.zero_()

    if device.type == "cpu":
        aggregate_cpu(X, G, F, Y)
    else:
        aggregate_gpu(X, G, F, Y)

    return Y


def broadcast(Y, G, F, X=None):
    device = Y.device
    if X is None:
        X = torch.zeros(
            G.shape + (Y.shape[-1],),
            device=device,
            dtype=Y.dtype
        )

    if device.type == "cpu":
        broadcast_cpu(Y, G, F, X)
    else:
        broadcast_gpu(Y, G, F, X)

    return X
    

def clustered_broadcast(Y, groups, counts, lengths, X=None):
    device = Y.device
    if X is None:
        X = torch.zeros(
            groups.shape + (Y.shape[-1],),
            device=device,
            dtype=Y.dtype
        )

    if device.type == "cpu":
        raise NotImplementedError
    else:
        N, H, C, E = Y.shape
        _, _, L, E = X.shape
   
        queries_per_block = min(L, 1024) 
        threads = queries_per_block
        blocks = (L//threads) + C + 1
        query_map = torch.ones((N, H, blocks),
                               dtype=torch.int32,
                               device=Y.device) * L 
        blocks_map = torch.ones_like(query_map,
                                     dtype=torch.int32,
                                     device=Y.device) * -1 
        _, sorted_group_indices = torch.sort(groups, descending=True, dim=-1)
        factors = torch.ones_like(counts, dtype=Y.dtype)
        clustered_broadcast_gpu(
            Y,
            groups,
            factors,
            X,
            lengths,
            blocks_map,
            query_map,
            counts,
            sorted_group_indices,
        )

    return X



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>
#


import os
import time
import unittest

import torch

from fast_transformers.attention import LinearAttention
from fast_transformers.masking import FullMask, LengthMask
from fast_transformers.recurrent.attention import RecurrentCrossLinearAttention


class TestRecurrentCrossLinearAttention(unittest.TestCase):
    def test_correctness(self):
        # Prepare the inputs
        N = 10
        H = 4
        E = 25
        M = 64
        L = 42
        S = 100
        q = torch.rand(N, L, H, E)
        k = torch.rand(N, S, H, E)
        v = torch.rand(N, S, H, M)
        m1 = FullMask(L, S)
        m2 = LengthMask(torch.full((N,), L, dtype=torch.int64))
        m3 = LengthMask(torch.full((N,), S, dtype=torch.int64))

        # Get the outputs from the attention in batch mode
        att = LinearAttention()
        att.eval()
        v_out1 = att(q, k, v, m1, m2, m3)

        # Get the output from the attention in recurrent mode
        att = RecurrentCrossLinearAttention()
        att.eval()
        v_out2_unstacked = []
        state = None
        for i in range(L):
            vi, state = att(q[:, i], k, v, m3, state=state)
            v_out2_unstacked.append(vi)
        v_out2 = torch.stack(v_out2_unstacked, dim=1)

        # Check that they match
        self.assertLess(torch.abs(v_out1 - v_out2).max(), 1e-6)


if __name__ == "__main__":
    unittest.main()




#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import torch

from .causal_product_cpu import causal_dot_product as causal_dot_product_cpu, \
    causal_dot_backward as causal_dot_backward_cpu

try:
    from .causal_product_cuda import \
        causal_dot_product as causal_dot_product_cuda, \
        causal_dot_backward as causal_dot_backward_cuda
except ImportError:
    causal_dot_product_cuda = causal_dot_backward_cuda = None


class CausalDotProduct(torch.autograd.Function):
    """Compute the weighted sum of values but attending only to previous
    values."""
    dot = {
        "cpu": causal_dot_product_cpu,
        "cuda": causal_dot_product_cuda
    }
    dot_backward = {
        "cpu": causal_dot_backward_cpu,
        "cuda": causal_dot_backward_cuda
    }

    @staticmethod
    def forward(ctx, Q, K, V):
        # Save the inputs for the gradient computation
        ctx.save_for_backward(Q, K, V)

        # Create the output tensor
        device = Q.device
        N, H, L, _ = Q.shape
        _, _, _, M = V.shape
        product = torch.zeros((N, H, L, M), device=device)

        # Actually perform the dot product
        CausalDotProduct.dot[device.type](
            Q.data,
            K.data,
            V.data,
            product
        )

        return product

    @staticmethod
    def backward(ctx, grad_out):
        # Extract the saved tensors
        Q, K, V = ctx.saved_tensors

        # Allocate memory for the gradients
        grad_Q = torch.zeros_like(Q)
        grad_K = torch.zeros_like(K)
        grad_V = torch.zeros_like(V)

        # Actually compute the gradients
        CausalDotProduct.dot_backward[Q.device.type](
            Q.data,
            K.data,
            V.data,
            grad_out,
            grad_Q,
            grad_K,
            grad_V
        )

        return grad_Q, grad_K, grad_V


# Alias the autograd functions to python style snake case naming
causal_dot_product = CausalDotProduct.apply



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

"""Implement transformer encoders and decoders as RNNs that will be used with
different recurrent attention mechanisms.

In all cases there exists no sequence dimension and the shapes are batch x
heads x dims.

This module's interface is designed with the linear attention in mind. The
interface is subject to change given the implementation of other recurrent
attentions.
"""

import warnings

import torch
from torch.nn import Dropout, LayerNorm, Linear, Module, ModuleList
import torch.nn.functional as F

from ._utils import check_state


class RecurrentTransformerEncoderLayer(Module):
    """Attention to the previous inputs and feed forward with skip connections.

    This transformer encoder layer is the recurrent dual of
    fast_transformers.transformers.TransformerEncoderLayer . The results should
    be identical given the same inputs and a lower triangular mask.

    Arguments
    ---------
        attention: The attention implementation to use given as a nn.Module
        d_model: The input feature dimensionality
        n_heads: The number of heads for the multi head attention
        d_ff: The dimensionality of the intermediate features after the
              attention (default: d_model*4)
        dropout: The dropout rate to apply to the intermediate features
                 (default: 0.1)
        activation: {'relu', 'gelu'} Which activation to use for the feed
                    forward part of the layer (default: relu)
    """
    def __init__(self, attention, d_model, n_heads, d_ff=None, dropout=0.1,
                 activation="relu"):
        super(RecurrentTransformerEncoderLayer, self).__init__()
        d_ff = d_ff or 4*d_model
        self.attention = attention
        self.linear1 = Linear(d_model, d_ff)
        self.linear2 = Linear(d_ff, d_model)
        self.norm1 = LayerNorm(d_model)
        self.norm2 = LayerNorm(d_model)
        self.dropout = Dropout(dropout)
        self.activation = F.relu if activation == "relu" else F.gelu

    def forward(self, x, state=None, memory=None):
        """Apply the transformer encoder to the input x using the provided
        memory.

        Arguments
        ---------
            x: The input features of shape (N, E) where N is the batch size and
               E is d_model passed in the constructor
            state: The state can vary depending on the attention implementation
            memory: **Deprecated** name for the state argument
        """
        # Normalize the state name
        state = check_state(state, memory)

        # Run the self attention and add it to the input
        x2, state = self.attention(x, x, x, state)
        x = x + self.dropout(x2)

        # Run the fully connected part of the layer
        y = x = self.norm1(x)
        y = self.dropout(self.activation(self.linear1(y)))
        y = self.dropout(self.linear2(y))

        return self.norm2(x+y), state


class RecurrentTransformerEncoder(Module):
    """RecurrentTransformerEncoder is a sequence of
    RecurrentTransformerEncoderLayer instances.

    RecurrentTransformerEncoder keeps a separate state per
    RecurrentTransformerEncoderLayer.

    Arguments
    ---------
        layers: list, RecurrentTransformerEncoderLayer instances or instances
                that implement the same interface
        norm_layer: A normalization layer to be applied to the final output
                    (default: None which means no normalization)
    """
    def __init__(self, layers, norm_layer=None):
        super(RecurrentTransformerEncoder, self).__init__()
        self.layers = ModuleList(layers)
        self.norm = norm_layer

    def forward(self, x, state=None, memory=None):
        """Apply all recurrent transformer layers to the input x using the
        provided state.

        Arguments
        ---------
            x: The input features of shape (N, E) where N is the batch size and
               E is d_model passed in the constructor of each recurrent
               transformer encoder layer
            state: A list of objects to be passed to each recurrent
                   transformer encoder layer
            memory: **Deprecated** name for the state argument
        """
        # Initialize the memory to None if not given
        state = check_state(state, memory)
        if state is None:
            state = [None]*len(self.layers)

        # Apply all the transformers
        for i, layer in enumerate(self.layers):
            x, s = layer(x, state[i])
            state[i] = s

        # Apply the normalization if needed
        if self.norm is not None:
            x = self.norm(x)

        return x, state


class RecurrentTransformerDecoderLayer(Module):
    """Attention to the previous inputs and a preprocessed memory.

    This transformer decoder layer is the recurrent dual of
    fast_transformers.transformers.TransformerDecoderLayer . The results should
    be identical given the same inputs and a lower triangular mask for x_mask.

    Arguments
    ---------
        self_attention: The attention implementation to use for self attention
                        given as a nn.Module
        cross_attention: The attention implementation to use for cross
                         attention given as a nn.Module
        d_model: The input feature dimensionality
        d_ff: The dimensionality of the intermediate features after the
              attention (default: d_model*4)
        dropout: The dropout rate to apply to the intermediate features
                 (default: 0.1)
        activation: {'relu', 'gelu'} Which activation to use for the feed
                    forward part of the layer (default: relu)
    """
    def __init__(self, self_attention, cross_attention, d_model, d_ff=None,
                 dropout=0.1, activation="relu"):
        super(RecurrentTransformerDecoderLayer, self).__init__()
        d_ff = d_ff or 4*d_model
        self.self_attention = self_attention
        self.cross_attention = cross_attention
        self.linear1 = Linear(d_model, d_ff)
        self.linear2 = Linear(d_ff, d_model)
        self.norm1 = LayerNorm(d_model)
        self.norm2 = LayerNorm(d_model)
        self.norm3 = LayerNorm(d_model)
        self.dropout = Dropout(dropout)
        self.activation = F.relu if activation == "relu" else F.gelu

    def forward(self, x, memory, memory_length_mask=None, state=None):
        """Apply the transformer decoder to the input x and also attend to
        memory.

        Note the memory mask is assumed to be a full mask.

        Arguments
        ---------
            x: The input features of shape (N, E) where N is the batch size and
               E is d_model passed in the constructor
            memory: A sequence of features (N, S, E) that the input will attend
                    to. S is the sequence length and E is the same as for x.
            memory_length_mask: An implementation of a BaseMask that encodes
                                how many elements each memory sequence in the
                                batch consists of.
            state: The state varies depending on the attention implementations
                   but it allows for recurrent implementation.
        """
        # Normalize the mask
        N = x.shape[0]
        L = memory.shape[1]
        memory_length_mask = memory_length_mask or \
            LengthMask(x.new_full((N,)), L, dtype=torch.int64)

        # Extract the individual states for the self attention and the cross
        # attention
        self_state, cross_state = state or [None, None]

        # First apply the self attention and add it to the input
        x2, self_state = self.self_attention(x, x, x, state=self_state)
        x = self.norm1(x + self.dropout(x2))

        # Secondly apply the cross attention and add it to the previous output
        x2, cross_state = self.cross_attention(
            x, memory, memory, memory_length_mask, state=cross_state
        )
        x = self.norm2(x + self.dropout(x2))

        # Finally run the fully connected part of the layer
        y = x
        y = self.dropout(self.activation(self.linear1(y)))
        y = self.dropout(self.linear2(y))

        return self.norm3(x+y), [self_state, cross_state]


class RecurrentTransformerDecoder(Module):
    """RecurrentTransformerDecoder is little more than a sequence of
    RecurrentTransformerDecoderLayer instances.

    Simlar to the recurrent encoder a separate state is kept per decoder layer.

    Arguments
    ---------
        layers: list, RecurrentTransformerDecoderLayer instances or instances
                that implement the same interface
        norm_layer: A normalization layer to be applied to the final output
                    (default: None which means no normalization)
    """
    def __init__(self, layers, norm_layer=None):
        super(RecurrentTransformerDecoder, self).__init__()
        self.layers = ModuleList(layers)
        self.norm = norm_layer

    def forward(self, x, memory, memory_length_mask=None, state=None):
        """Apply all recurrent transformer layers to the input x using the
        provided state.

        Arguments
        ---------
            x: The input features of shape (N, E) where N is the batch size and
               E is d_model passed in the constructor
            memory: A sequence of features (N, S, E) that the input will attend
                    to. S is the sequence length and E is the same as for x.
            memory_length_mask: An implementation of a BaseMask that encodes
                                how many elements each memory sequence in the
                                batch consists of
            state: A list of objects to be passed to each recurrent
                   transformer decoder layer
        """
        # Initialize the state to None if not given
        if state is None:
            state = [None]*len(self.layers)

        # Apply all the transformers
        for i, layer in enumerate(self.layers):
            x, s = layer(x, memory, memory_length_mask=memory_length_mask,
                         state=state[i])
            state[i] = s

        # Apply the normalization if needed
        if self.norm is not None:
            x = self.norm(x)

        return x, state



#
# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by Angelos Katharopoulos <angelos.katharopoulos@idiap.ch>,
# Apoorv Vyas <avyas@idiap.ch>
#

import os
import time
import unittest

import torch

from fast_transformers.sparse_product import sparse_dot_product


class TestSparseProductCUDA(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        if not torch.cuda.is_available():
            raise unittest.SkipTest("No CUDA capable device detected")

    def test_single_query(self):
        X = torch.randn(1, 1, 1, 32).cuda()
        Y = torch.randn(1, 1, 100, 32).cuda()
        lengths = torch.full((1,), 1, dtype=torch.int32).cuda()
        topk = (torch.cumsum(torch.rand(1, 1, 1, 10)*10, dim=-1)).long().cuda()

        products = sparse_dot_product(
            X,
            Y,
            topk,
        )
        all_products = torch.einsum("nhle,nhse->nhls", X, Y)

        self.assertLess(
            torch.max(torch.abs(
                products.squeeze() -
                all_products[0, 0, 0, topk[0, 0, 0]]
           )),
            1e-4
        )

    def test_simple_product(self):
        X = torch.randn(10, 4, 100, 32).cuda()
        Y = torch.randn(10, 4, 100, 32).cuda()
        lengths = torch.full((10,), 100, dtype=torch.int32).cuda()
        topk = (torch.cumsum(torch.rand(10, 4, 100, 10)*10, dim=-1)).long().cuda()

        A = torch.randn(10, 4, 100, 100).to(X.device).requires_grad_(False) 
        topk_v, topk = torch.topk(A, 10, dim=-1)
        topk = topk.contiguous()

        products = sparse_dot_product(
            X,
            Y,
            topk,
        )
        all_products = torch.einsum("nhle,nhse->nhls", X, Y)

        self.assertLess(
            torch.max(torch.abs(
                products -
                all_products[
                    torch.arange(10).view(10, 1, 1, 1),
                    torch.arange(4).view(1, 4, 1, 1),
                    torch.arange(100).view(1, 1, 100, 1),
                    topk
                ]
           )),
            1e-4
        )

    @unittest.skipUnless(os.getenv("BENCHMARK_TESTS", ""), "no benchmarks")
    def test_small_benchmark(self):
        N = 12
        H = 8
        L = 1000
        S = 1000
        E = 32
        k = 32
        X = torch.randn(N, H, L, E).cuda()
        Y = torch.randn(N, H, S, E).cuda()

        A = torch.randn(N, H, L, S).to(X.device).requires_grad_(False) 
        topk_v, topk = torch.topk(A, k, dim=-1)
        topk = topk.contiguous()

        for i in range(1000):
            products = sparse_dot_product(
                X,
                Y,
                topk,
            )
        torch.cuda.synchronize()
        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        products = sparse_dot_product(
            X,
            Y,
            topk,
        )
        e.record()
        torch.cuda.synchronize()
        t_s = s.elapsed_time(e)
        for i in range(1000):
            torch.einsum("nhle,nhse->nhls", X, Y)
        s = torch.cuda.Event(enable_timing=True)
        e = torch.cuda.Event(enable_timing=True)
        s.record()
        torch.einsum("nhle,nhse->nhls", X, Y)
        e.record()
        torch.cuda.synchronize()
        t_f = s.elapsed_time(e)
        print("Sparse: {}, Full: {}, F/S: {}".format(t_s, t_f, t_f/t_s))


if __name__ == "__main__":
    unittest.main()
